from handyscope.library import libtiepie
import ctypes


class OscilloscopeChannel:
    """Class for an oscilloscope channel.

    Attributes:
        CONNECTOR_TYPES (dict): dict which maps connector types as strs to 
                                their libtiepie int version
        COUPLINGS (dict): dict which maps couplings as strs to their libtiepie
                          int version
        TRIGGER_KINDS (dict): dict which maps trigger kinds as strs to their 
                              libtiepie int version
        TRIGGER_CONDITIONS (dict): dict which maps trigger conditions as strs
                                   to their libtiepie int version
    """
    # See also Generator.CONNECTOR_TYPES
    CONNECTOR_TYPES = {"unknown":   0,
                       "BNC":       1,
                       "Banana":    2,
                       "Powerplug": 4}

    COUPLINGS = {"unknown":  0,
                 "V DC":     1,
                 "V AC":     2,
                 "A DC":     4,
                 "A AC":     8,
                 "Ohm":     16}

    # Note: OscilloscopeChannel.TRIGGER_KINDS is the same as 
    # TriggerInput.TRIGGER_KINDS, but defined seperately here for perspicuity.
    # In the API documentation there's the additional define "TKM_NONE", which
    # has the same value as the - according to the docs also valid - define 
    # "TK_UNKNOWN". Thus it is ignored here.
    TRIGGER_KINDS = {"unknown": 0,
                     "rising": 1,
                     "falling": 2,
                     "in window": 4,
                     "out window": 8,
                     "any": 16,
                     "enter window": 32,
                     "exit window": 64,
                     "pulsewidth positive": 128,
                     "pulsewidth negative": 256}

    TRIGGER_CONDITIONS = {"unknown": 0,
                          "none":    1,
                          "smaller": 2,
                          "larger":  4}

    TRIGGER_LVL_MODES = {"unknown":  0,
                         "relative": 1,
                         "absolute": 2}

    RAW_DATA_TYPES = {"unknown":    0,
                      "int8":       1,
                      "int16":      2,
                      "int32":      4,
                      "int64":      8,
                      "uint8":     16,
                      "uint16":    32,
                      "uint32":    64,
                      "uint64":   128,
                      "float32":  256,
                      "float64":  512}

    def __init__(self, dev_handle, ch_idx):
        """Constructor for an oscilloscope channel.

        Args:
            dev_handle (int): Libtiepie device handle
            ch_idx (int): Channel index
        """
        self._dev_handle = dev_handle
        self._idx = ch_idx

    @property
    def bandwidths_available(self):
        """Get available input bandwidths in Hz.

        Returns:
            tuple: Available input bandwidths in Hz.
        """
        # Get length of list
        ranges_len = libtiepie.ScpChGetBandwidths(self._dev_handle,
                                                  self._idx, None, 0)

        # Initialize double array
        bandwidths = (ctypes.c_double * ranges_len)()

        # Write the actual data to the array
        libtiepie.ScpChGetBandwidths(self._dev_handle, self._idx,
                                     ctypes.byref(bandwidths), ranges_len)

        # Convert to a normal python list
        bandwidths = tuple(bandwidths)

        return bandwidths

    @property
    def bandwidth(self):
        """Get or set the input bandwidth."""
        return libtiepie.ScpChGetBandwidth(self._dev_handle, self._idx)

    @bandwidth.setter
    def bandwidth(self, value):
        libtiepie.ScpChSetBandwidth(self._dev_handle, self._idx, value)

    @property
    def is_safe_ground_available(self):
        """Check whether the channel has a safe ground.

        Returns:
            bool: True if the channel has a safe ground.
        """
        return libtiepie.ScpChHasSafeGround(self._dev_handle, self._idx) == 1

    @property
    def safe_ground_enabled(self):
        """Get or set whether the safe ground is enabled.
        
        Not tested.
        """
        return libtiepie.ScpChGetSafeGroundEnabled(self._dev_handle,
                                                   self._idx) == 1

    @safe_ground_enabled.setter
    def safe_ground_enabled(self, value):
        libtiepie.ScpChSetSafeGroundEnabled(self._dev_handle, self._idx, value)

    @property
    def safe_ground_threshold_min(self):
        """Get the minimum safe ground threshold current.
        
        Not tested.
        """
        return libtiepie.ScpChGetSafeGroundThresholdMin(self._dev_handle,
                                                        self._idx)

    @property
    def safe_ground_threshold_max(self):
        """Get the maximum safe ground threshold current.
        
        Not tested.
        """
        return libtiepie.ScpChGetSafeGroundThresholdMax(self._dev_handle,
                                                        self._idx)

    @property
    def safe_ground_threshold(self):
        """Get or set the safe ground threshold current.
        
        Not tested.
        """
        return libtiepie.ScpChGetSafeGroundThreshold(self._dev_handle,
                                                     self._idx)

    @safe_ground_threshold.setter
    def safe_ground_threshold(self, value):
        libtiepie.ScpChSetSafeGroundThreshold(self._dev_handle, self._idx,
                                              value)

    def verify_safe_ground_threshold(self, threshold):
        """Verify a safe ground threshold without setting it in the hardware.

        Not tested.
        
        Args:
            threshold (float): The safe ground threshold to verify.
        Returns:
            float: The safe ground threshold the hardware would set.
                   (The hardware might not set the safe ground threshold
                    due to clipping.)
        """
        return libtiepie.ScpChVerifySafeGroundThreshold(self._dev_handle, self._idx,
                                                        threshold)

    @property
    def connector_type(self):
        """Get the connector type.

        Returns:
            str: Connector type (key of :py:attr:`handyscope.oscilloscopeChannel.OscilloscopeChannel.CONNECTOR_TYPES`)
        """
        raw_type = libtiepie.ScpChGetConnectorType(self._dev_handle, self._idx)
        for key in self.CONNECTOR_TYPES:
            if raw_type == self.CONNECTOR_TYPES[key]:
                return key

        raise ValueError("Unknown connector type: %d" % raw_type)

    @property
    def is_differential(self):
        """Check if the channel is differential.

        Returns:
            bool: True if channel is differential, False otherwise.
        """
        return libtiepie.ScpChIsDifferential(self._dev_handle, self._idx) == 1

    @property
    def impedance(self):
        """Get the channel's input impedance.

        Returns:
            float: The channel's input impedance
        """
        return libtiepie.ScpChGetImpedance(self._dev_handle, self._idx)

    @property
    def couplings_available(self):
        """Get available couplings.

        Returns:
            tuple: Available couplings (keys of :py:attr:`handyscope.oscilloscopeChannel.OscilloscopeChannel.COUPLINGS`)
        """
        raw_couplings = libtiepie.ScpChGetCouplings(
            self._dev_handle, self._idx)
        _couplings = []

        # if no couplings are available, return unknown
        if raw_couplings == self.COUPLINGS["unknown"]:
            _couplings.append("unknown")
        # else do a detailed analysis...
        else:
            # ... by iterating over every possible coupling...
            for key in self.COUPLINGS:
                # ... and ignoring "unknown" (already handled above)
                if key == "unknown":
                    pass
                elif raw_couplings & self.COUPLINGS[key] == self.COUPLINGS[key]:
                    _couplings.append(key)

        return tuple(_couplings)

    @property
    def coupling(self):
        """Get or set the current coupling (key of
        :py:attr:`handyscope.oscilloscopeChannel.OscilloscopeChannel.COUPLINGS`)."""
        raw_coupling = libtiepie.ScpChGetCoupling(self._dev_handle, self._idx)
        for key in self.COUPLINGS:
            if raw_coupling == self.COUPLINGS[key]:
                return key

        raise ValueError("Unknown coupling: %d" % raw_coupling)

    @coupling.setter
    def coupling(self, value):
        libtiepie.ScpChSetCoupling(
            self._dev_handle, self._idx, self.COUPLINGS[value])

    @property
    def is_enabled(self):
        """Get or set if the channel is enabled."""
        return libtiepie.ScpChGetEnabled(self._dev_handle, self._idx) == 1

    @is_enabled.setter
    def is_enabled(self, value):
        libtiepie.ScpChSetEnabled(self._dev_handle, self._idx, value)

    @property
    def probe_gain(self):
        """Get or set the probe gain."""
        return libtiepie.ScpChGetProbeGain(self._dev_handle, self._idx)

    @probe_gain.setter
    def probe_gain(self, value):
        libtiepie.ScpChSetProbeGain(self._dev_handle, self._idx, value)

    @property
    def probe_offset(self):
        """Get or set the probe offset in V."""
        return libtiepie.ScpChGetProbeOffset(self._dev_handle, self._idx)

    @probe_offset.setter
    def probe_offset(self, value):
        libtiepie.ScpChSetProbeOffset(self._dev_handle, self._idx, value)

    @property
    def is_auto_range(self):
        """Get or set if autoranging is enabled."""
        return libtiepie.ScpChGetAutoRanging(self._dev_handle, self._idx) == 1

    @is_auto_range.setter
    def is_auto_range(self, value):
        libtiepie.ScpChSetAutoRanging(self._dev_handle, self._idx, value)

    @property
    def ranges_available(self):
        """Get available ranges in V.

        Returns:
            tuple: Available ranges in V.
        """
        # Get length of list
        ranges_len = libtiepie.ScpChGetRanges(
            self._dev_handle, self._idx, None, 0)

        # Initialize double array
        ranges = (ctypes.c_double * ranges_len)()

        # Write the actual data to the array
        libtiepie.ScpChGetRanges(
            self._dev_handle, self._idx, ctypes.byref(ranges), ranges_len)

        # Convert to a normal python list
        ranges = tuple(ranges)

        return ranges

    @property
    def range(self):
        """Get or set the current range in V."""
        return libtiepie.ScpChGetRange(self._dev_handle, self._idx)

    @range.setter
    def range(self, value):
        libtiepie.ScpChSetRange(self._dev_handle, self._idx, value)

    @property
    def is_max_range_reachable(self):
        """Check whether the maximum of the range is reachable.

        Returns:
            bool: True, if the maximum of the range is reachable,
                  False otherwise.
        """
        return libtiepie.ScpChIsRangeMaxReachable(self._dev_handle, self._idx) == 1

    @property
    def is_trig_enabled(self):
        """Get or set if the channel trigger is enabled."""
        return libtiepie.ScpChTrGetEnabled(self._dev_handle, self._idx) == 1

    @is_trig_enabled.setter
    def is_trig_enabled(self, value):
        libtiepie.ScpChTrSetEnabled(self._dev_handle, self._idx, value)

    @property
    def trig_kinds_available(self):
        """Get available trigger kinds.

        Returns:
            tuple: Available trigger kinds (keys of
                   :py:attr:`handyscope.oscilloscopeChannel.OscilloscopeChannel.TRIGGER_KINDS`)
        """
        raw_kinds = libtiepie.ScpChTrGetKinds(self._dev_handle, self._idx)
        _kinds = []

        # If no trigger kinds are available, return unknown
        if raw_kinds == self.TRIGGER_KINDS["unknown"]:
            _kinds.append("unknown")
        # Else do a detailed analysis...
        else:
            # ...by iterating over every possible kind ...
            for key in self.TRIGGER_KINDS:
                # ...and ignoring "unknown" (already handled above)
                if key == "unknown":
                    pass
                elif raw_kinds & self.TRIGGER_KINDS[key] == self.TRIGGER_KINDS[key]:
                    _kinds.append(key)

        return tuple(_kinds)

    @property
    def trig_kind(self):
        """Get or set the current trigger kind (key of
        :py:attr:`handyscope.oscilloscopeChannel.OscilloscopeChannel.TRIGGER_KINDS`)"""
        raw_kind = libtiepie.ScpChTrGetKind(self._dev_handle, self._idx)
        for key in self.TRIGGER_KINDS:
            if raw_kind == self.TRIGGER_KINDS[key]:
                return key

        raise ValueError("Unknown trigger kind: %d" % raw_kind)

    @trig_kind.setter
    def trig_kind(self, value):
        libtiepie.ScpChTrSetKind(
            self._dev_handle, self._idx, self.TRIGGER_KINDS[value])

    @property
    def trig_lvl_cnt(self):
        """Get the count of trigger levels.

        Returns:
            int: Count of trigger levels.
        """
        return libtiepie.ScpChTrGetLevelCount(self._dev_handle, self._idx)

    @property
    def trig_lvl(self):
        """Get or set the trigger levels.

        Value range depends on
        :py:attr:`handyscope.oscilloscopeChannel.OscilloscopeChannel.trig_lvl_mode`.
        """
        _lvls = []
        for idx in range(self.trig_lvl_cnt):
            _lvls.append(libtiepie.ScpChTrGetLevel(
                self._dev_handle, self._idx, idx))
        return tuple(_lvls)

    @trig_lvl.setter
    def trig_lvl(self, iterable):
        for idx, value in enumerate(iterable):
            libtiepie.ScpChTrSetLevel(self._dev_handle, self._idx, idx, value)

    @property
    def trig_lvl_modes_available(self):
        """Get the available trigger level modes.

        Returns:
            tuple: Available trigger level modes (keys of
                   :py:attr:`handyscope.oscilloscopeChannel.OscilloscopeChannel.TRIGGER_LVL_MODES`)
        """
        raw_modes = libtiepie.ScpChTrGetLevelModes(self._dev_handle, self._idx)
        _modes = []

        # If no trigger level modes are available, return unknown
        if raw_modes == self.TRIGGER_LVL_MODES["unknown"]:
            _modes.append("unknown")
        # Else do a detailed analysis...
        else:
            # ...by iterating over every possible kind ...
            for key in self.TRIGGER_LVL_MODES:
                # ...and ignoring "unknown" (already handled above)
                if key == "unknown":
                    pass
                elif raw_modes & self.TRIGGER_LVL_MODES[key] == self.TRIGGER_LVL_MODES[key]:
                    _modes.append(key)

        return tuple(_modes)

    @property
    def trig_lvl_mode(self):
        """Get or set the current trigger level mode.

        Possible trigger level modes are the keys of
        :py:attr:`handyscope.oscilloscopeChannel.OscilloscopeChannel.TRIGGER_LVL_MODES`.
        """
        raw_mode = libtiepie.ScpChTrGetLevelMode(self._dev_handle, self._idx)
        for key, value in self.TRIGGER_LVL_MODES.items():
            if raw_mode == value:
                return key

        raise ValueError("Unknown trigger level mode: %d" % raw_mode)

    @trig_lvl_mode.setter
    def trig_lvl_mode(self, value):
        libtiepie.ScpChTrSetLevelMode(
            self._dev_handle, self._idx, self.TRIGGER_LVL_MODES[value])

    @property
    def trig_hysteresis_cnt(self):
        """Get the count of trigger hysteresises.

        Returns:
            int: Count of trigger hysteresises
        """
        return libtiepie.ScpChTrGetHysteresisCount(self._dev_handle, self._idx)

    @property
    def trig_hysteresis(self):
        """Get or set the trigger hysteresis."""
        _hyst = []
        for idx in range(self.trig_hysteresis_cnt):
            _hyst.append(libtiepie.ScpChTrGetHysteresis(
                self._dev_handle, self._idx, idx))

        return tuple(_hyst)

    @trig_hysteresis.setter
    def trig_hysteresis(self, iterable):
        for idx, value in enumerate(iterable):
            libtiepie.ScpChTrSetHysteresis(
                self._dev_handle, self._idx, idx, value)

    @property
    def trig_conditions_available(self):
        """Get the available trigger conditions.

        Returns:
            tuple: Available trigger conditions (keys of
                   :py:attr:`handyscope.oscilloscopeChannel.OscilloscopeChannel.TRIGGER_CONDITIONS`)
        """
        raw_conds = libtiepie.ScpChTrGetConditions(self._dev_handle, self._idx)
        _conds = []

        # If no trigger conditions are available, return unknown
        if raw_conds == self.TRIGGER_CONDITIONS["unknown"]:
            _conds.append("unknown")
        # Else do a detailed analysis...
        else:
            # ...by iterating over every possible kind ...
            for key in self.TRIGGER_CONDITIONS:
                # ...and ignoring "unknown" (already handled above)
                if key == "unknown":
                    pass
                elif raw_conds & self.TRIGGER_CONDITIONS[key] == self.TRIGGER_CONDITIONS[key]:
                    _conds.append(key)

        return tuple(_conds)

    @property
    def trig_condition(self):
        """Get or set the current trigger condition (key of
        :py:attr:`handyscope.oscilloscopeChannel.OscilloscopeChannel.TRIGGER_CONDITIONS`)"""
        raw_cond = libtiepie.ScpChTrGetCondition(self._dev_handle, self._idx)
        for key, value in self.TRIGGER_CONDITIONS.items():
            if raw_cond == value:
                return key

        raise ValueError("Unknown trigger condition: %d" % raw_cond)

    @trig_condition.setter
    def trig_condition(self, value):
        libtiepie.ScpChTrSetCondition(
            self._dev_handle, self._idx, self.TRIGGER_CONDITIONS[value])

    @property
    def trig_time_cnt(self):
        """Get the count of trigger times.

        Returns:
            int: Count of trigger times
        """
        return libtiepie.ScpChTrGetTimeCount(self._dev_handle, self._idx)

    @property
    def trig_time(self):
        """Get or set the trigger times."""
        times = []

        for idx in range(self.trig_time_cnt):
            times.append(libtiepie.ScpChTrGetTime(
                self._dev_handle, self._idx, idx))

        return tuple(times)

    @trig_time.setter
    def trig_time(self, iterable):
        for idx, value in enumerate(iterable):
            libtiepie.ScpChTrSetTime(self._dev_handle, self._idx, idx, value)

    def verify_trig_time(self, trigger_times):
        """Verify trigger times without setting it in the hardware.

        Args:
            trigger_times (iterable): The trigger times to verify.
        Returns:
            float: The trigger times the hardware would set.
                   (The hardware might not set the trigger times due to
                   clipping.)
        """
        set_trigger_times = []
        for idx, value in enumerate(trigger_times):
            trigger_time = libtiepie.ScpChTrVerifyTime(self._dev_handle,
                                                       self._idx, idx, value)
            set_trigger_times.append(trigger_time)
        return set_trigger_times

    @property
    def is_trig_available(self):
        """Check if trigger is available.

        Returns:
            bool: True if trigger is available, False otherwise.
        """
        # TODO Check if both function calls are necessary. To be on the safe 
        # side, currently both are called. Check if trigger support is given 
        # under the currently selected measure mode
        measure_mode_ok = libtiepie.ScpChHasTrigger(
            self._dev_handle, self._idx) == 1
        # Check if trigger is available with current settings
        settings_ok = libtiepie.ScpChTrIsAvailable(
            self._dev_handle, self._idx) == 1

        return measure_mode_ok & settings_ok

    @property
    def is_triggered(self):
        """Check if the channel has been triggered.

        Returns:
            bool: True, if the channel has been triggered, False otherwise.
        """
        return libtiepie.ScpChTrIsTriggered(self._dev_handle, self._idx) == 1

    @property
    def is_available(self):
        """Check if channel is available.

        Returns:
            bool: True if channel is available, False otherwise.
        """
        return libtiepie.ScpChIsAvailable(self._dev_handle, self._idx) == 1

    @property
    def is_connection_test_available(self):
        """Check if connection test is available.

        Returns:
            bool: True if connection test is available, False otherwise.
        """
        return libtiepie.ScpChHasConnectionTest(self._dev_handle, self._idx) == 1

    @property
    def data_range(self):
        """Get the minimum and maximum values of the input range of the measured data.

        Returns:
            tuple: Minimum and maximum values of the input range as float
        """
        # Init floats
        range_min = ctypes.c_double()
        range_max = ctypes.c_double()

        # Get data
        libtiepie.ScpChGetDataValueRange(
            self._dev_handle, self._idx, ctypes.byref(range_min),
            ctypes.byref(range_max))

        # Get values and return them
        return range_min.value, range_max.value

    @property
    def data_value_min(self):
        """Get the minimum value of the input range of the measured data.

        Returns:
            float: Minimum value of the input range
        """
        return libtiepie.ScpChGetDataValueMin(self._dev_handle, self._idx)

    @property
    def data_value_max(self):
        """Get the maximum value of the input range of the measured data.

        Returns:
            float: Maximum value of the input range
        """
        return libtiepie.ScpChGetDataValueMax(self._dev_handle, self._idx)

    @property
    def raw_data_type(self):
        """Get the raw type of measured data.

        Returns:
            str: Raw data type of the measured data.
        """
        raw_type = libtiepie.ScpChGetDataRawType(self._dev_handle, self._idx)
        for key, value in self.RAW_DATA_TYPES.items():
            if raw_type == value:
                return key
        return key

    @property
    def raw_data_range(self):
        """Get the minimum, zero and maximum raw value of the input range
        of the measured data.

        Returns:
            tuple: Minimum, zero, maximum raw values.
        """
        minimum = ctypes.c_uint64()
        zero = ctypes.c_uint64()
        maximum = ctypes.c_uint64()

        libtiepie.ScpChGetDataRawValueRange(self._dev_handle, self._idx,
                                            ctypes.byref(minimum),
                                            ctypes.byref(zero),
                                            ctypes.byref(maximum))

        return minimum.value, zero.value, maximum.value

    @property
    def raw_value_min(self):
        """Get the minimum raw value for the input range of the measured data.

        Returns:
            int: Minimum raw value for the input range.
        """
        return libtiepie.ScpChGetDataRawValueMin(self._dev_handle, self._idx)

    @property
    def raw_value_max(self):
        """Get the maximum raw value of the input range of the measured data.

        Returns:
            int: Maximum raw value of the input range.
        """
        return libtiepie.ScpChGetDataRawValueMax(self._dev_handle, self._idx)

    @property
    def raw_value_zero(self):
        """Get the raw value which corresponds to zero for the input range
        of the measured data.

        Returns:
            int: Raw value which corresponds to zero for the input range.
        """
        return libtiepie.ScpChGetDataRawValueZero(self._dev_handle, self._idx)



from handyscope.library import libtiepie
from handyscope.device import Device
import ctypes


class Generator(Device):
    """Class for a generator.

    Attributes:
        CONNECTOR_TYPES (dict): dict which maps connector types as strs to
                                their libtiepie int version
        GENERATOR_STATUSES (dict): dict which maps generator statuses as strs
                                   to their libtiepie int version
        SIGNAL_TYPES (dict): dict which maps signal types as strs to
                             their libtiepie int version
        FREQUENCY_MODES (dict): dict which maps frequency modes as strs to
                                their libtiepie int version
        GENERATOR_MODES (dict): dict which maps generator modes as strs to
                                their libtiepie int version
    """
    # See also OscilloscopeChannel.CONNECTOR_TYPES
    CONNECTOR_TYPES = {"unknown":   0,
                       "BNC":       1,
                       "Banana":    2,
                       "Powerplug": 4}

    GENERATOR_STATUSES = {"unknown":      0,
                          "stopped":      1,
                          "running":      2,
                          "burst active": 4,
                          "waiting":      8}

    SIGNAL_TYPES = {"unknown":      0,
                    "sine":         1,
                    "triangle":     2,
                    "square":       4,
                    "DC":           8,
                    "noise":       16,
                    "arbitrary":   32,
                    "pulse":       64}

    FREQUENCY_MODES = {"unknown":   0,
                       "signal":    1,
                       "sample":    2}

    GENERATOR_MODES = {"unknown":                          0,
                       "continuous":                       1,
                       "burst count":                      2,
                       "gated periods":                    4,
                       "gated":                            8,
                       "gated period start":              16,
                       "gated period finish":             32,
                       "gated run":                       64,
                       "gated run output":               128,
                       "burst sample count":             256,
                       "burst sample count output":      512,
                       "burst segment count":           1024,
                       "burst segment count output":    2048}

    RAW_DATA_TYPES = {"unknown":    0,
                      "int8":       1,
                      "int16":      2,
                      "int32":      4,
                      "int64":      8,
                      "uint8":     16,
                      "uint16":    32,
                      "uint32":    64,
                      "uint64":   128,
                      "float32":  256,
                      "float64":  512}

    _device_type = "Gen"

    def __init__(self, instr_id, id_kind="product id"):
        """Contructor for a generator.

        Args:
            instr_id (int or str): Device list index, product ID
                                   (listed in dict PRODUCT_IDS) or serial number
            id_kind (str): the kind of the given instr_id
                           (listed in dict ID_KINDS)
        """
        super().__init__(instr_id, id_kind, self._device_type)

    @property
    def connector_type(self):
        """Get the connector type.

        Returns:
            str: The connector type, key of
                 :py:attr:`handyscope.generator.Generator.CONNECTOR_TYPES`
        """
        raw_type = libtiepie.GenGetConnectorType(self._dev_handle)
        for key, value in self.CONNECTOR_TYPES.items():
            if raw_type == value:
                return key

        raise ValueError("Unknown connector type: %d" % raw_type)

    @property
    def is_differential(self):
        """Check if output is differential.

        Returns:
            bool: True if differential, False otherwise.
        """
        return libtiepie.GenIsDifferential(self._dev_handle) == 1

    @property
    def impedance(self):
        """Get output impedance.

        Returns:
            float: Output impedance in Ohm.
        """
        return libtiepie.GenGetImpedance(self._dev_handle)

    @property
    def resolution(self):
        """Get DAC resolution.

        Returns:
            int: Resolution of the DAC in bits.
        """
        return libtiepie.GenGetResolution(self._dev_handle)

    @property
    def out_min(self):
        """Get minimum available output voltage.

        Returns:
            float: Minimum available output voltage in Volt.
        """
        return libtiepie.GenGetOutputValueMin(self._dev_handle)

    @property
    def out_max(self):
        """Get maximum available output voltage.

        Returns:
            float: Maximum available output voltage in Volt.
        """
        return libtiepie.GenGetOutputValueMax(self._dev_handle)

    @property
    def is_controllable(self):
        """Check if the generator is controllable.

        Returns:
            bool: True if controllable, False otherwise.
        """
        return libtiepie.GenIsControllable(self._dev_handle) == 1

    @property
    def status(self):
        """Get the current generator status.

        Returns:
            str: Generator status, key of
                 :py:attr:`handyscope.generator.Generator.GENERATOR_STATUSES`
        """
        raw_status = libtiepie.GenGetStatus(self._dev_handle)

        for key, value in self.GENERATOR_STATUSES.items():
            if raw_status == value:
                return key

        raise ValueError("Unknown generator status: %d" % raw_status)

    @property
    def is_out_on(self):
        """Get or set if the generator output is enabled."""
        return libtiepie.GenGetOutputOn(self._dev_handle) == 1

    @is_out_on.setter
    def is_out_on(self, value):
        libtiepie.GenSetOutputOn(self._dev_handle, value)

    @property
    def is_out_inv_available(self):
        """Get whether the generator output is invertible."""
        return libtiepie.GenHasOutputInvert(self._dev_handle) == 1

    @property
    def is_out_inv(self):
        """Get or set if the generator output is inverted."""
        return libtiepie.GenGetOutputInvert(self._dev_handle) == 1

    @is_out_inv.setter
    def is_out_inv(self, value):
        libtiepie.GenSetOutputInvert(self._dev_handle, value)

    def start(self):
        """Start the generator.

        Returns:
            bool: True if successful, False otherwise.
        """
        return libtiepie.GenStart(self._dev_handle) == 1

    def stop(self):
        """Stop the generator.

        Returns:
            bool: True if successful, False otherwise.
        """
        return libtiepie.GenStop(self._dev_handle) == 1

    @property
    def is_running(self):
        """Check whether the generator is running."""
        return libtiepie.GenIsRunning(self._dev_handle) == 1

    @property
    def signal_types_available(self):
        """Get available signal types.

        Returns:
            tuple: Available signal types, keys of
                   :py:attr:`handyscope.generator.Generator.SIGNAL_TYPES`
        """
        raw_types = libtiepie.GenGetSignalTypes(self._dev_handle)
        _types = []

        if raw_types == self.SIGNAL_TYPES["unknown"]:
            _types.append("unknown")
        else:
            for key, value in self.SIGNAL_TYPES.items():
                if key == "unknown":
                    pass
                elif raw_types & value == value:
                    _types.append(key)

        return tuple(_types)

    @property
    def signal_type(self):
        """Get or set the currently active signal type."""
        raw_type = libtiepie.GenGetSignalType(self._dev_handle)

        for key, value in self.SIGNAL_TYPES.items():
            if raw_type == value:
                return key

        raise ValueError("Unknown signal type: %d" % raw_type)

    @signal_type.setter
    def signal_type(self, value):
        libtiepie.GenSetSignalType(self._dev_handle, self.SIGNAL_TYPES[value])

    @property
    def is_amplitude_available(self):
        """Get whether setting the amplitude is available."""
        return libtiepie.GenHasAmplitude(self._dev_handle) == 1

    @property
    def amplitude_min(self):
        """Get the minimum available amplitude.

        Returns:
            float: Minimum available amplitude in Volt.
        """
        return libtiepie.GenGetAmplitudeMin(self._dev_handle)

    @property
    def amplitude_max(self):
        """Get the maximum available amplitude.

        Returns:
            float: Maximum available amplitude in Volt.
        """
        return libtiepie.GenGetAmplitudeMax(self._dev_handle)

    @property
    def amplitude(self):
        """Get or set the amplitude in Volt."""
        return libtiepie.GenGetAmplitude(self._dev_handle)

    @amplitude.setter
    def amplitude(self, value):
        libtiepie.GenSetAmplitude(self._dev_handle, value)

    @property
    def amplitude_ranges_available(self):
        """Get the available amplitude ranges.

        Returns:
            tuple: Available amplitude ranges as floats in Volt.
        """
        list_len = libtiepie.GenGetAmplitudeRanges(self._dev_handle, None, 0)

        buffer = (ctypes.c_double * list_len)()

        libtiepie.GenGetAmplitudeRanges(
            self._dev_handle, ctypes.byref(buffer), list_len)

        return tuple(buffer)

    @property
    def amplitude_range(self):
        """Get or set the currently used amplitude range."""
        return libtiepie.GenGetAmplitudeRange(self._dev_handle)

    @amplitude_range.setter
    def amplitude_range(self, value):
        libtiepie.GenSetAmplitudeRange(self._dev_handle, value)

    @property
    def is_amplitude_autorange(self):
        """Get or set if amplitude autoranging is enabled."""
        return libtiepie.GenGetAmplitudeAutoRanging(self._dev_handle) == 1

    @is_amplitude_autorange.setter
    def is_amplitude_autorange(self, value):
        libtiepie.GenSetAmplitudeAutoRanging(self._dev_handle, value)

    def verify_amplitude(self, amplitude):
        """Verify an amplitude without setting it in the hardware.

        Args:
            amplitude (float): The amplitude to verify.
        Returns:
            float: The amplitude the hardware would set.
                   (The hardware might not set the amplitude
                    due to clipping.)
        """
        return libtiepie.GenVerifyAmplitude(self._dev_handle, amplitude)

    @property
    def is_offset_available(self):
        """Get whether setting the offset is available."""
        return libtiepie.GenHasOffset(self._dev_handle) == 1

    @property
    def offset_min(self):
        """Get the minimum available offset.

        Returns:
            float: Minimum available offset in Volt.
        """
        return libtiepie.GenGetOffsetMin(self._dev_handle)

    @property
    def offset_max(self):
        """Get the maximum available offset.

        Returns:
            float: Maximum available offset in Volt.
        """
        return libtiepie.GenGetOffsetMax(self._dev_handle)

    @property
    def offset(self):
        """Get or set the current offset in Volt."""
        return libtiepie.GenGetOffset(self._dev_handle)

    @offset.setter
    def offset(self, value):
        libtiepie.GenSetOffset(self._dev_handle, value)

    def verify_offset(self, offset):
        """Verify an offset without setting it in the hardware.

        Args:
            offset (float): The offset to verify.
        Returns:
            float: The offset the hardware would set.
                   (The hardware might not set the offset
                    due to clipping.)
        """
        return libtiepie.GenVerifyOffset(self._dev_handle, offset)

    @property
    def is_frequency_available(self):
        """Get whether setting the frequency is available."""
        return libtiepie.GenHasFrequency(self._dev_handle) == 1

    @property
    def freq_min(self):
        """Get the minimum available frequency.

        Returns:
            float: Minimum available frequency in Hz.
        """
        return libtiepie.GenGetFrequencyMin(self._dev_handle)

    @property
    def freq_max(self):
        """Get the maximum available frequency.

        Returns:
            float: Maximum available frequency in Hz.
        """
        return libtiepie.GenGetFrequencyMax(self._dev_handle)

    @property
    def freq(self):
        """Get or set the currently used frequency in Hz."""
        return libtiepie.GenGetFrequency(self._dev_handle)

    @freq.setter
    def freq(self, value):
        libtiepie.GenSetFrequency(self._dev_handle, value)

    def verify_frequency(self, frequency):
        """Verify a frequency without setting it in the hardware.

        Args:
            frequency (float): The frequency to verify.
        Returns:
            float: The frequency the hardware would set.
                   (The hardware might not set the frequency
                    due to clipping.)
        """
        return libtiepie.GenVerifyFrequency(self._dev_handle, frequency)

    @property
    def freq_modes_available(self):
        """Get the available frequency modes.

        Returns:
            tuple: Available frequency modes, keys of
                   :py:attr:`handyscope.generator.Generator.FREQUENCY_MODES`
        """
        raw_modes = libtiepie.GenGetFrequencyModes(self._dev_handle)
        _modes = []

        if raw_modes == self.FREQUENCY_MODES["unknown"]:
            _modes.append("unknown")
        else:
            for key, value in self.FREQUENCY_MODES.items():
                if key == "unknown":
                    pass
                elif raw_modes & value == value:
                    _modes.append(key)

        return tuple(_modes)

    @property
    def freq_mode(self):
        """Get or set the currently active frequency mode."""
        raw_mode = libtiepie.GenGetFrequencyMode(self._dev_handle)

        for key, value in self.FREQUENCY_MODES.items():
            if raw_mode == value:
                return key

        raise ValueError("Unknown frequency mode: %d" % raw_mode)

    @freq_mode.setter
    def freq_mode(self, value):
        libtiepie.GenSetFrequencyMode(
            self._dev_handle, self.FREQUENCY_MODES[value])

    @property
    def is_phase_available(self):
        """Get whether setting the phase is available."""
        return libtiepie.GenHasPhase(self._dev_handle) == 1

    @property
    def phase_min(self):
        """Get the minimum available signal phase.

        Returns:
            float: Minimum available signal phase in degree.
        """
        return libtiepie.GenGetPhaseMin(self._dev_handle) * 360

    @property
    def phase_max(self):
        """Get the maximum available signal phase.

        Returns:
            float: Maximum available signal phase in degree.
        """
        return libtiepie.GenGetPhaseMax(self._dev_handle) * 360

    @property
    def phase(self):
        """Get or set the signal phase in degree."""
        return libtiepie.GenGetPhase(self._dev_handle) * 360

    @phase.setter
    def phase(self, value):
        libtiepie.GenSetPhase(self._dev_handle, value/360)

    def verify_phase(self, phase):
        """Verify a phase without setting it in the hardware.

        Args:
            phase (float): The phase to verify.
        Returns:
            float: The phase the hardware would set.
                   (The hardware might not set the phase
                    due to clipping.)
        """
        return libtiepie.GenVerifyPhase(self._dev_handle, phase/360) * 360

    @property
    def is_symmetry_available(self):
        """Get whether setting the symmetry is available."""
        return libtiepie.GenHasSymmetry(self._dev_handle) == 1

    @property
    def symmetry_min(self):
        """Get the minimum available symmetry.

        The symmetry of a signal defines the ratio between the length of
        positive part of a period and the length of the negative part of
        a period of the generated signal.

        Returns:
            float: Minimum available symmetry, value between 0 and 1.
        """
        return libtiepie.GenGetSymmetryMin(self._dev_handle)

    @property
    def symmetry_max(self):
        """Get the maximum available symmetry.

        The symmetry of a signal defines the ratio between the length of
        positive part of a period and the length of the negative part of a
        period of the generated signal.

        Returns:
            float: Maximum available symmetry, value between 0 and 1.
        """
        return libtiepie.GenGetSymmetryMax(self._dev_handle)

    @property
    def symmetry(self):
        """Get or set the maximum available symmetry.

        The symmetry of a signal defines the ratio between the length of
        positive part of a period and the length of the negative part of
        a period of the generated signal.
        """
        return libtiepie.GenGetSymmetry(self._dev_handle)

    @symmetry.setter
    def symmetry(self, value):
        libtiepie.GenSetSymmetry(self._dev_handle, value)

    def verify_symmetry(self, symmetry):
        """Verify a symmetry without setting it in the hardware.

        Args:
            symmetry (float): The symmetry to verify.
        Returns:
            float: The symmetry the hardware would set.
                   (The hardware might not set the symmetry
                    due to clipping.)
        """
        return libtiepie.GenVerifySymmetry(self._dev_handle, symmetry)

    @property
    def is_pulse_width_available(self):
        """Get whether setting the pulse width is available."""
        return libtiepie.GenHasWidth(self._dev_handle) == 1

    @property
    def pulse_width_min(self):
        """Get the minimum available pulse width.

        Available for signal type "pulse".

        Returns:
            float: Minimum available pulse width in seconds.
        """
        return libtiepie.GenGetWidthMin(self._dev_handle)

    @property
    def pulse_width_max(self):
        """Get the maximum available pulse width.

        Available for signal type "pulse".

        Returns:
            float: Maximum available pulse width in seconds.
        """
        return libtiepie.GenGetWidthMax(self._dev_handle)

    @property
    def pulse_width(self):
        """Get or set the pulse width in seconds.

        Available for signal type "pulse".
        """
        return libtiepie.GenGetWidth(self._dev_handle)

    @pulse_width.setter
    def pulse_width(self, value):
        libtiepie.GenSetWidth(self._dev_handle, value)

    def verify_pulse_width(self, pulse_width):
        """Verify a pulse width without setting it in the hardware.

        Args:
            pulse_width (float): The pulse width to verify.
        Returns:
            float: The pulse width the hardware would set.
                   (The hardware might not set the pulse width
                    due to clipping.)
        """
        return libtiepie.GenVerifyWidth(self._dev_handle, pulse_width)

    @property
    def leading_edge_time_min(self):
        """Get the minimum available leading edge time.

        Available for signal type "pulse".
        
        Not tested.
        """
        return libtiepie.GenGetLeadingEdgeTimeMin(self._dev_handle)

    @property
    def leading_edge_time_max(self):
        """Get the maximum available leading edge time.

        Available for signal type "pulse".
        
        Not tested.
        """
        return libtiepie.GenGetLeadingEdgeTimeMax(self._dev_handle)

    @property
    def leading_edge_time(self):
        """Get or set the available leading edge time.

        Available for signal type "pulse".
        
        Not tested.
        """
        return libtiepie.GenGetLeadingEdgeTime(self._dev_handle)

    @leading_edge_time.setter
    def leading_edge_time(self, value):
        libtiepie.GenSetLeadingEdgeTime(self._dev_handle, value)

    def verify_leading_edge_time(self, leading_edge_time):
        """Verify a leading edge time without setting it in the hardware.

        Available for signal type "pulse".
        
        Not tested.

        Args:
            leading_edge_time (float): The leading edge time to verify.
        Returns:
            float: The leading edge time the hardware would set.
                   (The hardware might not set the leading edge time
                    due to clipping.)
        """
        return libtiepie.GenVerifyLeadingEdgeTime(self._dev_handle,
                                                  leading_edge_time)

    @property
    def trailing_edge_time_min(self):
        """Get the minimum available trailing edge time.

        Available for signal type "pulse".
        
        Not tested.
        """
        return libtiepie.GenGetTrailingEdgeTimeMin(self._dev_handle)

    @property
    def trailing_edge_time_max(self):
        """Get the maximum available trailing edge time.

        Available for signal type "pulse".
        
        Not tested.
        """
        return libtiepie.GenGetTrailingEdgeTimeMax(self._dev_handle)

    @property
    def trailing_edge_time(self):
        """Get or set the available trailing edge time.

        Available for signal type "pulse".
        
        Not tested.
        """
        return libtiepie.GenGetTrailingEdgeTime(self._dev_handle)

    @trailing_edge_time.setter
    def trailing_edge_time(self, value):
        libtiepie.GenSetTrailingEdgeTime(self._dev_handle, value)

    def verify_trailing_edge_time(self, trailing_edge_time):
        """Verify a trailing edge time without setting it in the hardware.

        Available for signal type "pulse".
        
        Not tested.

        Args:
            trailing_edge_time (float): The trailing edge time to verify.
        Returns:
            float: The trailing edge time the hardware would set.
                   (The hardware might not set the trailing edge time
                    due to clipping.)
        """
        return libtiepie.GenVerifyTrailingEdgeTime(self._dev_handle,
                                                   trailing_edge_time)

    @property
    def is_arb_data_available(self):
        """Get whether the current signal type supports controlling the
        arbitrary waveform buffer.
        
        Not tested.
        """
        return libtiepie.GenHasData(self._dev_handle) == 1

    @property
    def arb_data_length_min(self):
        """Get the minimum length (number of samples) of arbitrary data.

        Available for signal type "arbitrary".

        Returns:
            int: Minimum length of arbitrary data
        """
        return libtiepie.GenGetDataLengthMin(self._dev_handle)

    @property
    def arb_data_length_max(self):
        """Get the maximum length (number of samples) of arbitrary data.

        Available for signal type "arbitrary".

        Returns:
            int: Maximum length of arbitrary data
        """
        return libtiepie.GenGetDataLengthMax(self._dev_handle)

    @property
    def arb_data_length(self):
        """Get the length of the currently loaded arbitrary data.

        Available for signal type "arbitrary".

        Returns:
            int: Length of currently loaded arbitrary data.
        """
        return libtiepie.GenGetDataLength(self._dev_handle)

    def verify_arb_data_length(self, arb_data_length):
        """Verify a length for arbitrary data without setting it in the
        hardware.

        Available for signal type "arbitrary".
        
        Not tested.

        Args:
            arb_data_length (int): The length to verify.
        Returns:
            int: The length the hardware would set.
                 (The hardware might not set the length
                 due to clipping.)
        """
        return libtiepie.GenVerifyDataLength(self._dev_handle,
                                             arb_data_length)

    def arb_data(self, value_list):
        """Fill the arbitrary data buffer of the generator.

        The values are normalized by the generator itself: The highest
        absolute value equals the set amplitude value, 0 corresponds to the
        set offset value. If value_list is empty, the buffer gets cleared.

        Args:
            value_list (list): List with arbitrary data samples
        """
        list_len = len(value_list)

        if list_len == 0:
            pointer = None
        else:
            buffer = (ctypes.c_float * list_len)(*value_list)
            pointer = ctypes.byref(buffer)

        libtiepie.GenSetData(self._dev_handle, pointer, list_len)

    @property
    def arb_raw_type(self):
        """Get the data type for setting raw arbitrary values."""
        raw_type = libtiepie.GenGetDataRawType(self._dev_handle)

        for key, value in self.RAW_DATA_TYPES.items():
            if raw_type == value:
                return key

    @property
    def arb_data_raw_range(self):
        """Get the minimum, zero and maximum raw value of the arbitrary data
        buffer range as int.

        Returns:
            tuple: Minimum, zero, maximum raw value for the arbitrary data
                   buffer range as int.
        """
        minimum = ctypes.c_uint64()
        zero = ctypes.c_uint64()
        maximum = ctypes.c_uint64()

        libtiepie.GenGetDataRawValueRange(self._dev_handle,
                                          ctypes.byref(minimum),
                                          ctypes.byref(zero),
                                          ctypes.byref(maximum))

        return minimum.value, zero.value, maximum.value

    @property
    def arb_data_raw_min(self):
        """Get the minimum raw value for the arbitrary data buffer which
        corresponds to the current amplitude value.

        Returns:
            int: Minimum raw value for the arbitrary data buffer.
        """
        return libtiepie.GenGetDataRawValueMin(self._dev_handle)

    @property
    def arb_data_raw_max(self):
        """Get the maximum raw value for the arbitrary data buffer which
        corresponds to the current amplitude value.

        Returns:
            int: Maximum raw value for the arbitrary data buffer.
        """
        return libtiepie.GenGetDataRawValueMax(self._dev_handle)

    @property
    def arb_data_raw_zero(self):
        """Get the raw value which corresponds to zero in the arbitrary data
        buffer.

        Returns:
            int: Raw value which corresponds to zero in the arbitrary
                 data buffer.
        """
        return libtiepie.GenGetDataRawValueZero(self._dev_handle)

    def arb_data_raw(self, value_list):
        """Fill the arbitrary data buffer of the generator with raw values.

        Args:
            value_list (list): List with arbitrary data samples
        """
        list_len = len(value_list)

        if list_len == 0:
            pointer = None
        else:
            buffer = (ctypes.c_float * list_len)(*value_list)
            pointer = ctypes.byref(buffer)

        libtiepie.GenSetDataRaw(self._dev_handle, pointer, list_len)

    @property
    def modes_native_available(self):
        """Get the natively available generator modes.

        The natively available modes are the built-in modes of the generetator,
        independent of current settings.

        Returns:
            tuple: Available modes, keys of
                  :py:attr:`handyscope.generator.Generator.GENERATOR_MODES`
        """
        raw_modes = libtiepie.GenGetModesNative(self._dev_handle)
        _modes = []

        if raw_modes == self.GENERATOR_MODES["unknown"]:
            _modes.append("unknown")
        else:
            for key, value in self.GENERATOR_MODES.items():
                if key == "unknown":
                    pass
                elif raw_modes & value == value:
                    _modes.append(key)

        return tuple(_modes)

    @property
    def modes_available(self):
        """Get the currently available generator modes.

        The currently available modes depend on the current generator settings.

        Returns:
            tuple: Available modes, keys of
                   :py:attr:`handyscope.generator.Generator.GENERATOR_MODES`
        """
        raw_modes = libtiepie.GenGetModes(self._dev_handle)
        _modes = []

        if raw_modes == self.GENERATOR_MODES["unknown"]:
            _modes.append("unknown")
        else:
            for key, value in self.GENERATOR_MODES.items():
                if key == "unknown":
                    pass
                elif raw_modes & value == value:
                    _modes.append(key)

        return tuple(_modes)

    @property
    def mode(self):
        """Get or set the current generator mode (keys of
        :py:attr:`handyscope.generator.Generator.GENERATOR_MODES`)."""
        raw_mode = libtiepie.GenGetMode(self._dev_handle)

        for key, value in self.GENERATOR_MODES.items():
            if raw_mode == value:
                return key

        raise ValueError("Unknown generator mode: %d" % raw_mode)

    @mode.setter
    def mode(self, value):
        libtiepie.GenSetMode(self._dev_handle, self.GENERATOR_MODES[value])

    @property
    def is_burst_active(self):
        """Check if a burst is active.

        Returns:
            bool: True if burst is active, False otherwise.
        """
        return libtiepie.GenIsBurstActive(self._dev_handle) == 1

    @property
    def burst_cnt_min(self):
        """Get the minimum available burst count.

        Available in generator burst modes.

        Returns:
            int: Minimum available burst count.
        """
        return libtiepie.GenGetBurstCountMin(self._dev_handle)

    @property
    def burst_cnt_max(self):
        """Get the maximum available burst count.

        Available in generator burst modes.

        Returns:
            int: Maximum available burst count.
        """
        return libtiepie.GenGetBurstCountMax(self._dev_handle)

    @property
    def burst_cnt(self):
        """Get or set the burst count.

        Available in generator burst modes.
        """
        return libtiepie.GenGetBurstCount(self._dev_handle)

    @burst_cnt.setter
    def burst_cnt(self, value):
        libtiepie.GenSetBurstCount(self._dev_handle, value)

    @property
    def burst_sample_cnt_min(self):
        """Get the minimum available burst sample count.

        Available in generator sample burst modes.

        Returns:
            int: Minimum available sample burst count.
        """
        return libtiepie.GenGetBurstSampleCountMin(self._dev_handle)

    @property
    def burst_sample_cnt_max(self):
        """Get the maximum available burst sample count.

        Available in generator sample burst modes.

        Returns:
            int: Maximum available sample burst count.
        """
        return libtiepie.GenGetBurstSampleCountMax(self._dev_handle)

    @property
    def burst_sample_cnt(self):
        """Get or set the burst sample count.

        Available in generator sample burst modes.
        """
        return libtiepie.GenGetBurstSampleCount(self._dev_handle)

    @burst_sample_cnt.setter
    def burst_sample_cnt(self, value):
        libtiepie.GenSetBurstSampleCount(self._dev_handle, value)

    @property
    def burst_segment_cnt_min(self):
        """Get the minimum available burst segment count.

        Available in generator segment burst modes.

        Returns:
            int: Minimum available segment burst count.
        """
        return libtiepie.GenGetBurstSegmentCountMin(self._dev_handle)

    @property
    def burst_segment_cnt_max(self):
        """Get the maximum available burst segment count.

        Available in generator segment burst modes.

        Returns:
            int: Maximum available segment burst count.
        """
        return libtiepie.GenGetBurstSegmentCountMax(self._dev_handle)

    @property
    def burst_segment_cnt(self):
        """Get or set the burst segment count.

        Available in generator segment burst modes.
        """
        return libtiepie.GenGetBurstSegmentCount(self._dev_handle)

    @burst_segment_cnt.setter
    def burst_segment_cnt(self, value):
        libtiepie.GenSetBurstSegmentCount(self._dev_handle, value)

    def verify_burst_segment_cnt(self, burst_segment_cnt):
        """Verify a burst segment count without setting it in the hardware.

        Not tested.
        
        Args:
            burst_segment_cnt (int): The burst segment count to verify.
            
        Returns:
            int: The burst segment count the hardware would set.
                 (The hardware might not set the burst segment count
                 due to clipping.)
        """
        return libtiepie.GenVerifyBurstSegmentCount(self._dev_handle,
                                                    burst_segment_cnt)



import ctypes
from handyscope.library import libtiepie


class TriggerInput:
    TRIGGER_KINDS = {"unknown": 0,
                     "rising": 1,
                     "falling": 2,
                     "in window": 4,
                     "out window": 8,
                     "any": 16,
                     "enter window": 32,
                     "exit window": 64,
                     "pulsewidth positive": 128,
                     "pulsewidth negative": 256}

    # See api doc, macro "TRIGGER_IO_ID"
    TRIGGER_IDS = {"EXT 1":                 0 << 24 | 3 << 20 | 1 << 8 | 0,
                   "EXT 2":                 0 << 24 | 3 << 20 | 2 << 8 | 0,
                   "EXT 3":                 0 << 24 | 3 << 20 | 3 << 8 | 0,
                   "Generator start":       0 << 24 | 2 << 20 | 0 << 8 | 0,
                   "Generator stop":        0 << 24 | 2 << 20 | 0 << 8 | 1,
                   "Generator new period":  0 << 24 | 2 << 20 | 0 << 8 | 2}

    def __init__(self, dev_handle, trig_in_idx):
        self._dev_handle = dev_handle
        self._idx = trig_in_idx

    @property
    def is_available(self):
        return libtiepie.DevTrInIsAvailable(self._dev_handle, self._idx) == 1

    @property
    def trigger_id(self):
        raw_id = libtiepie.DevTrInGetId(self._dev_handle, self._idx)
        for key in self.TRIGGER_IDS:
            if self.TRIGGER_IDS[key] == raw_id:
                return key

        raise ValueError("Unknown trigger input id %d" % raw_id)

    @property
    def name(self):
        # get length of name string
        str_len = libtiepie.DevTrInGetName(
            self._dev_handle, self._idx, None, 0)

        # initialize mutable string buffer
        str_buffer = ctypes.create_string_buffer(str_len)

        # write the actual name to the buffer
        libtiepie.DevTrInGetName(
            self._dev_handle, self._idx, str_buffer, str_len)

        # convert to a normal python string
        name = str_buffer.value.decode('utf-8')

        return name

    @property
    def kinds_available(self):
        raw_kinds = libtiepie.DevTrInGetKinds(self._dev_handle, self._idx)
        _kinds = []

        # If no trigger kinds are available, return unknown
        if raw_kinds == self.TRIGGER_KINDS["unknown"]:
            _kinds.append("unknown")
        # Else do a detailed analysis...
        else:
            # ...by iterating over every possible kind ...
            for key in self.TRIGGER_KINDS:
                # ...and ignoring "unknown" (already handled above)
                if key == "unknown":
                    pass
                elif raw_kinds & self.TRIGGER_KINDS[key] == self.TRIGGER_KINDS[key]:
                    _kinds.append(key)

        return _kinds

    @property
    def kind(self):
        raw_kind = libtiepie.DevTrInGetKind(self._dev_handle, self._idx)
        for key in self.TRIGGER_KINDS:
            if raw_kind == self.TRIGGER_KINDS[key]:
                return key

        raise ValueError("Unknown trigger kind: %d" % raw_kind)

    @kind.setter
    def kind(self, value):
        libtiepie.DevTrInSetKind(
            self._dev_handle, self._idx, self.TRIGGER_KINDS[value])

    @property
    def is_enabled(self):
        return libtiepie.DevTrInGetEnabled(self._dev_handle, self._idx) == 1

    @is_enabled.setter
    def is_enabled(self, value):
        libtiepie.DevTrInSetEnabled(self._dev_handle, self._idx, value)

    @property
    def is_triggered(self):
        """Check if the oscilloscope trigger input caused a trigger.

        Only valid for oscilloscope trigger inputs, otherwise always False
        is returned.

        Returns:
            bool: True if trigger input belongs to an oscilloscope and is
                  triggered, false otherwise.
        """
        try:
            result = libtiepie.ScpTrInIsTriggered(
                self._dev_handle, self._idx) == 1
        except IOError as err:
            if str(err) == "[-3]: INVALID_HANDLE":
                result = False
            else:
                raise

        return result



from handyscope.library import libtiepie
from handyscope.device import Device
import ctypes


class I2CHost(Device):
    """Class for an I2CHost.

    An I2CHost is the I2C bus master included in Handyscopes. It is accessible
    via external connectors (D-Sub).
    """
    _device_type = "I2C"

    def __init__(self, instr_id, id_kind="product id"):
        """Constructor for an I2C host.

        Args:
            instr_id (int or str): Device list index, product ID (listed
                                   in dict PRODUCT_IDS) or serial number
            id_kind (str): the kind of the given instr_id (listed in dict
                           ID_KINDS)
        """
        super().__init__(instr_id, id_kind, self._device_type)

    @property
    def clock_freq_max(self):
        """Get the maximum available clock frequency of the I2C clock line.

        Returns:
            float: maximum clock frequency in Hz
        """
        return libtiepie.I2CGetSpeedMax(self._dev_handle)

    @property
    def clock_freq(self):
        """Get or set the current I2C clock frequency in Hz."""
        return libtiepie.I2CGetSpeed(self._dev_handle)

    @clock_freq.setter
    def clock_freq(self, value):
        libtiepie.I2CSetSpeed(self._dev_handle, value)

    def verify_clock_freq(self, clock_freq):
        """Verify a clock frequency without setting it in the
        hardware.

        Args:
            clock_freq (float): The clock frequency to verify.
        Returns:
            float: The clock frequency the hardware would set.
                 (The hardware might not set the clock frequency
                 due to clipping.)
        """
        return libtiepie.I2CVerifySpeed(self._dev_handle,
                                        clock_freq)

    def is_internal_address(self, address):
        """Check if given address is used internally.

        Args:
            address (int): I2C address

        Returns:
            bool: True, if given address is internally used; False otherwise.
        """
        return libtiepie.I2CIsInternalAddress(self._dev_handle, address) == 1

    @property
    def internal_adresses(self):
        """Get internally used addresses.

        Returns:
            tuple: Used adresses.
        """
        # Get length of list
        adresses_len = libtiepie.I2CGetInternalAddresses(self._dev_handle,
                                                         None, 0)

        # Initialize double array
        adresses = (ctypes.c_uint16 * adresses_len)()

        # Write the actual data to the array
        libtiepie.I2CGetInternalAddresses(self._dev_handle,
                                          ctypes.byref(adresses),
                                          adresses_len)

        # Convert to a normal python list
        adresses = tuple(adresses)

        return adresses

    def read(self, address, no_bytes, send_stop=True):
        """Read the given number of bytes from the given address.

        Args:
            address   (int):  I2C address
            no_bytes  (int):  Number of bytes to read
            send_stop (bool): Whether to send a stop bit (defaults to True).

        Returns:
            list of int: List with the received bytes.
        """
        buffer = (ctypes.c_uint8 * no_bytes)()
        libtiepie.I2CRead(self._dev_handle, address, ctypes.byref(buffer),
                          no_bytes, send_stop)
        return list(buffer)

    def read_byte(self, address):
        """Read one byte from the given address.

        Args:
            address (int): I2C address

        Returns:
            int: The received byte.
        """
        buffer = ctypes.c_uint8()
        libtiepie.I2CReadByte(self._dev_handle, address, ctypes.byref(buffer))
        return buffer.value

    def read_word(self, address):
        """Read one word from the given address.

        Args:
            address (int): I2C address

        Returns:
            int: The received word.
        """
        buffer = ctypes.c_uint16()
        libtiepie.I2CReadWord(self._dev_handle, address, ctypes.byref(buffer))
        return buffer.value

    def write(self, address, data, send_stop=True):
        """Write the given data to the address.

        Args:
            address     (int):          I2C address
            data        (list of int):  List of bytes to be written
            send_stop   (bool):         Whether to send a stop bit (defaults to
                                        True).

        Returns:
            bool: True if write succeeded, False otherwise.
        """
        data_len = len(data)
        buffer = (ctypes.c_uint8 * data_len)(*data)
        result = libtiepie.I2CWrite(self._dev_handle, address,
                                    ctypes.byref(buffer), data_len, send_stop)

        return result == 1

    def write_read(self, address, data, no_bytes):
        """Write and read data to/from the address.

        Args:
            address (int): I2C address.
            data (list of int): List of bytes to be written
            no_bytes  (int):  Number of bytes to read
        """
        data_len = len(data)
        write_buffer = (ctypes.c_uint8 * data_len)(*data)
        read_buffer = (ctypes.c_ubyte * no_bytes)()
        libtiepie.I2CWriteRead(self._dev_handle, address,
                               ctypes.byref(write_buffer),
                               data_len, ctypes.byref(read_buffer))
        return list(read_buffer)

    def write_byte(self, address, data_byte):
        """Write the given byte to the address.

        Args:
            address   (int): I2C address
            data_byte (int): Data byte to be written

        Returns:
            bool: True if write succeeded, False otherwise.
        """
        result = libtiepie.I2CWriteByte(self._dev_handle, address, data_byte)

        return result == 1

    def write_byte_byte(self, address, data_byte1, data_byte2):
        """Write the two given bytes to the address.

        Args:
            address    (int): I2C address
            data_byte1 (int): First byte to be written
            data_byte2 (int): Second byte to be written

        Returns:
            bool: True if write succeeded, False otherwise.
        """
        result = libtiepie.I2CWriteByteByte(self._dev_handle, address,
                                            data_byte1, data_byte2)

        return result == 1

    def write_word(self, address, data_word):
        """Write the given word to the address.

        Args:
            address   (int): I2C address
            data_word (int): Data word to write

        Returns:
            bool: True if write succeeded, False otherwise.
        """
        result = libtiepie.I2CWriteWord(self._dev_handle, address, data_word)

        return result == 1

    def write_byte_word(self, address, data_byte, data_word):
        """Write the given byte and the given word to the address.

        Args:
            address   (int): I2C address
            data_byte (int): Data byte to write
            data_word (int): Data word to write

        Returns:
            bool: True if write succeeded, False otherwise.
        """
        result = libtiepie.I2CWriteByteWord(self._dev_handle, address,
                                            data_byte, data_word)

        return result == 1

    def scan(self):
        """Scan for available nodes on the bus.

        Returns:
            tuple of int: valid addresses
        """
        valid_addresses = []

        # Only check allowed addresses: "Two groups of eight addresses
        # (0000 XXX and 1111 XXX) are reserved"
        # `see official I2C-bus specification and user manual
        # <http://www.nxp.com/documents/user_manual/UM10204.pdf>`_
        for address in range(0x08, 0x77):
            if self.is_internal_address(address):
                continue
            else:
                try:
                    self.write(address, [0])
                except OSError as err:
                    # If no ACK was received, there is no device listening on
                    # this address
                    if err.args[0] == "[-15]: NO_ACKNOWLEDGE":
                        continue
                # If code didn't raise an exception, i.e. ACK was received,
                # address is valid!
                valid_addresses.append(address)

        return tuple(valid_addresses)



import time
import numpy as np
import json
import os, inspect
from scipy.signal import gausspulse


def _intersample_peak(signal):
    """Calculate position of maximum of the given signal.

    Args:
        signal: The signal to use.

    Returns:
        float: Position of the maximum.
    """
    signal = np.absolute(signal)
    pos_max = np.argmax(signal)
    pos_max += 0.5 * (signal[pos_max - 1] - signal[pos_max + 1]) / \
               (signal[pos_max - 1] - 2 * signal[pos_max] + signal[pos_max + 1])
    return pos_max


def _load_sync_offset_config():
    """Load the known sync offset config.

    Returns:
        dict: Sync offset to Handyscopes.
    """
    path = os.path.dirname(os.path.abspath(inspect.stack()[0][1]))
    if os.name == 'nt':
        path += '\\'
    elif os.name == 'posix':
        path += '/'
    else:
        raise ValueError("Unknown operating system found")
    with open(path + 'config.json', 'r') as cfg_file:
        config = json.load(cfg_file)
    return config


def _measurement(gen, osz):
    """Take a single measurement on channel 1 and channel 2.

    Returns:
        list: Data of channel 1 and channel 2.
    """
    osz.start()
    gen.start()
    while not osz.is_data_ready:
        time.sleep(0.05)
    data = osz.retrieve_ch1_ch2()
    gen.stop()
    return data


def _save_sync_offset_config(new_config):
    """Save the new sync offset config.

    Args:
        new_config: The new config to save.
    """
    path = os.path.dirname(os.path.abspath(inspect.stack()[0][1]))
    if os.name == 'nt':
        path += '\\'
    elif os.name == 'posix':
        path += '/'
    else:
        raise ValueError("Unknown operating system found")
    with open(path + 'config.json', 'w') as cfg_file:
        json.dump(new_config, cfg_file, sort_keys=True, indent=4)


def calculate_sync_offset(gen, osz):
    """Calculate the sync offset between channel 1 and channel 2 for the
    current sample frequency of the oscilloscope and save it to file.
    Calculation is done with a center frequency of 1 MHz, because there is no
    real difference for reasonable center frequencies.

    Args:
        gen: Generator to use.
        osz: Oscilloscope to use.

    Returns:
        float: The calculated sync offset.

    Raises:
        AttributeError: If user aborts the process.
        ValueError: If user gives invalid input.
    """
    print(
        "No sync offset found in the database for the given device and sample"
        "frequency. Need to calculate it now. \n"
        "To do this the output channel of the handyscope has to be connected to"
        "both channel 1 and channel 2. \n"
        "Please change the assembly to look like this. Be careful to use BNC "
        "cables with the same length.")
    inp = input("Choices: \n"
                "c to continue the process, after you finished to rearrange the assembly \n"
                "a to abort the whole process")
    if inp == 'a':
        raise AttributeError("No snyc offset found and calculation was aborted")
    elif inp == 'c':
        # calculate input signal for generator
        t_c = gausspulse("cutoff", 1e6, 1.1, tpr=-40)
        steps = np.arange(-2 * t_c, 2 * t_c, 1 / osz.sample_freq)
        signal = gausspulse(steps, 1e6, 1.1)

        # settings for calculation and save old ones to restore them later
        old_gen_sig_type = gen.signal_type
        gen.signal_type = "arbitrary"
        old_gen_mode = gen.mode
        gen.mode = "burst count"
        old_gen_out_on = gen.is_out_on
        gen.is_out_on = True
        old_gen_burst_cnt = gen.burst_cnt
        gen.burst_cnt = 1
        gen.freq_mode = 'sample'
        old_gen_freq = gen.freq
        gen.freq = osz.sample_freq
        old_gen_amplitude = gen.amplitude
        gen.amplitude = 12
        old_gen_arb_data = gen.arb_data
        gen.arb_data(signal)
        old_osz_res = osz.resolution
        osz.resolution = 12
        old_osz_trig_timeout = osz.trig_timeout
        osz.trig_timeout = 5
        old_osz_ch0_range = osz.channels[0].range
        osz.channels[0].range = 20
        old_osz_ch1_range = osz.channels[1].range
        osz.channels[1].range = 20
        old_osz_ch0_trig_en = osz.channels[0].is_trig_enabled
        osz.channels[0].is_trig_enabled = 0
        old_osz_ch1_trig_en = osz.channels[1].is_trig_enabled
        osz.channels[1].is_trig_enabled = 0
        old_osz_trig_ins_3_en = osz.trig_ins[3].is_enabled
        osz.trig_ins[3].is_enabled = True
        old_osz_record_length = osz.record_length
        osz.record_length = signal.size

        def max_position():
            # retrieve data
            pos_max_ch1 = 0
            pos_max_ch2 = 0
            for mes in range(0, int(1e4)):
                data = _measurement(gen, osz)
                pos_max_ch1 += _intersample_peak(data[0])
                pos_max_ch2 += _intersample_peak(data[1])
                if mes % 500 == 0:
                    print("{} % done".format(int(mes / 100)))
            print("100 % done")

            # calculate sync offset
            pos_max_ch1 /= 1e4
            pos_max_ch2 /= 1e4
            offset = pos_max_ch1 - pos_max_ch2
            return round(offset, 2)

        sync_offset = max_position()

        # save the newly calculated sync offset
        current = _load_sync_offset_config()
        try:
            current[str(osz.serial_no)][str(osz.sample_freq)] = sync_offset
        except KeyError:  # no values for this serial are available yet
            current[str(osz.serial_no)] = {str(osz.sample_freq): sync_offset}
        _save_sync_offset_config(current)

        # restore old settings
        gen.signal_type = old_gen_sig_type
        gen.burst_cnt = old_gen_burst_cnt
        gen.mode = old_gen_mode
        gen.is_out_on = old_gen_out_on
        gen.freq = old_gen_freq
        gen.amplitude = old_gen_amplitude
        gen.arb_data = old_gen_arb_data
        osz.resolution = old_osz_res
        osz.trig_timeout = old_osz_trig_timeout
        osz.channels[0].range = old_osz_ch0_range
        osz.channels[1].range = old_osz_ch1_range
        osz.channels[0].is_trig_enabled = old_osz_ch0_trig_en
        osz.channels[1].is_trig_enabled = old_osz_ch1_trig_en
        osz.trig_ins[3].is_enabled = old_osz_trig_ins_3_en
        osz.record_length = old_osz_record_length

        print(
            "Sync offset was successfully calculated and saved for future use. "
            "Some variables of generator and oscilloscope were changed in the "
            "process, but should be restored. Eventually this was not done "
            "completely, so be carefully with the next results. \n"
            "Also consider uploading the new config file, for future "
            "consistency and easy use for other users of the same Handyscope.")
        inp = input("Choices: \n"
                    "c to continue with the normal measurement, after you "
                    "rearranged the assembly back\n"
                    "a to abort the measurement, to adjust the settings.")
        if inp == 'a':
            raise AttributeError("Measurement aborted by user.")
        elif inp == 'c':
            return sync_offset
        else:
            raise ValueError("Invalid option given")
    else:
        raise ValueError("Invalid option given")


def get_sync_offset(gen, osz):
    """Get the sync offset for the given serial number and current sample
    frequency of the oscilloscope. If unknown a new one will be calculated.

    Args:
        gen: The generator to use.
        osz: The oscilloscope to use.

    Returns:
        float: The sync offset of the oscilloscope.
    """
    config = _load_sync_offset_config()
    try:
        return config[str(osz.serial_no)][str(osz.sample_freq)]
    except KeyError:  # no sync offset known
        return calculate_sync_offset(gen, osz)


def measurement_jitter_free(gen, osz, gen_signal, n_avg, pause=0.1):
    """Take a measurement and calculate jitter free signal for channel 2. On
    channel 1 the input signal gets measured and on channel 2 the real
    measurement signal. This then gets corrected using the correlation between
    both channels, which results in jitter free values for channel 2.

    Args:
        gen: The generator to use.
        osz: The oscilloscope to use.
        gen_signal: The output signal of the generator.
        n_avg (int): Amount of measurements to be taken.
        pause (float): Pause between measurements in seconds.

    Returns:
        list: Values of channel 1.
        list: Values of channel 2.
        list: Jitter free values of channel 2.
    """
    ch_1 = []
    ch_2 = []
    jit_free = []
    # get the correct sync offset
    sync_offset = get_sync_offset(gen, osz)
    for mes in range(0, n_avg):
        # take one measurement
        data = _measurement(gen, osz)
        # calculate jitter free signal
        max_pos_x_0 = _intersample_peak(gen_signal)
        max_pos_channel_1 = _intersample_peak(data[0])
        # calculate difference and use the specific sync offset
        diff = max_pos_x_0 - max_pos_channel_1 + sync_offset
        # jitter free signal with 1d interpolation
        jitter_free_cur = np.interp(np.arange(len(data[1])) + diff,
                                    np.arange(len(data[1])), data[1])
        # append the current data to the overall
        ch_1.append(np.asanyarray(data[0]))
        ch_2.append(np.asanyarray(data[1]))
        jit_free.append(jitter_free_cur)
        time.sleep(pause)

    return ch_1, ch_2, jit_free



from handyscope.triggerInput import TriggerInput
from handyscope.triggerOutput import TriggerOutput


def test_firmware_ver(device):
    ver_list = device.firmware_ver.split('.')
    assert len(ver_list) == 4
    for element in ver_list:
        assert element.isdigit()


def test_calibration_date(device):
    import datetime

    assert type(device.calibration_date) is datetime.date


def test_serial_no(device):
    assert type(device.serial_no) is int
    assert device.serial_no >= 0


def test_product_id(device):
    from handyscope.deviceList import DeviceList

    product_id = device.product_id
    assert type(product_id) is str
    # Check if it is a valid key
    assert product_id in DeviceList.PRODUCT_IDS


def test_device_type(device):
    from handyscope.deviceList import DeviceList

    device_type = device.device_type
    assert type(device_type) is str
    # Check if it is a valid key
    assert device_type in DeviceList.DEVICE_TYPES


def test_short_name(device):
    from handyscope.deviceList import DeviceList

    name = device.short_name
    assert type(name) is str
    # For "Osc" and "I2C": Short name corresponds to product names -> look it up in the dict.
    # Attention: This is just an assumption made while working with the HS5 and libtiepie 0.6.3.
    if device.device_type in ["Osc", "I2C"]:
        assert name in DeviceList.PRODUCT_IDS
    # For "Gen": Short name consist out of "AWG " + product name. Attention: This is just an assumption made while
    # working with the HS5 and libtiepie 0.5 / 0.6.3.
        if device.device_type == "Gen":
            assert name in ["AWG " + product for product in DeviceList.PRODUCT_IDS]


def test_name(device):
    name = device.name
    name_list = name.split('-')
    assert len(name_list) == 2
    # first part of name is the short name
    assert name_list[0] == device.short_name


def test_long_name(device):
    name = device.long_name
    name_list = name.split(' ')
    assert len(name_list) == 2
    # first part of long name is always the manufacturer name
    assert name_list[0] == "Handyscope"
    # second part of long name is the name
    assert name_list[1] == device.name


def test_is_battery_available(device):
    assert device.is_battery_available == False


def test_is_removed(device):
    assert device.is_removed is False


def test_trig_in_cnt(device):
    assert type(device.trig_in_cnt) is int


def test_trig_ins(device):
    assert len(device.trig_ins) is device.trig_in_cnt
    for trig_in in device.trig_ins:
        assert type(trig_in) is TriggerInput


def test_trig_in_id(device):
    for trig_in_id in TriggerInput.TRIGGER_IDS:
        try:
            index = device.trig_in_by_id(trig_in_id)
            # If the id is supported by the device, a valid integer has to be returned. If not, 0xffff is returned.
            assert (index in range(device.trig_in_cnt)) | (index == 0xffff)
        except OSError as err:
            # If the device has no trigger inputs, an OSError is raised.
            assert str(err) == "[-2]: NOT_SUPPORTED"


def test_trig_out_cnt(device):
    assert type(device.trig_out_cnt) is int


def test_trig_outs(device):
    assert len(device.trig_outs) is device.trig_out_cnt
    for trig_out in device.trig_outs:
        assert type(trig_out) is TriggerOutput


def test_trig_out_id(device):
    for trig_out_id in TriggerOutput.TRIGGER_IDS:
        try:
            index = device.trig_out_by_id(trig_out_id)
            # If the id is supported by the device, a valid integer has to be returned. If not, 0xffff is returned.
            assert (index in range(device.trig_out_cnt)) | (index == 0xffff)
        except OSError as err:
            # If the device has no trigger outputs, an OSError is raised.
            assert str(err) == "[-2]: NOT_SUPPORTED"



from setuptools import setup
from os import path

here = path.abspath(path.dirname(__file__))
with open(path.join(here, 'README.rst')) as readme_file:
    long_description = readme_file.read()

setup(
    name='handyscope',

    description='Handyscope provides a Python interface to the mobile USB-oscilloscopes made by TiePie.',
    long_description=long_description,

    url='http://emt.uni-paderborn.de',
    author='Tim Hetkaemper',
    author_email='timh1@mail.uni-paderborn.de',
    license='BSD',

    # Automatically generate version number from git tags
    use_scm_version=True,

    packages=[
        'handyscope'
    ],

    # Additional data
    package_data={
        'handyscope': ['bin/libtiepie32.dll',
                       'bin/libtiepie64.dll',
                       'bin/libtiepie.so',
                       'config.json'],
    },

    # Runtime dependencies
    install_requires=[
        'numpy',
        'scipy'
    ],

    # Setup/build dependencies; setuptools_scm required for git-based versioning
    setup_requires=['setuptools_scm', 'pytest-runner'],

    # Test dependencies
    tests_require=['pytest'],

    # For a list of valid classifiers, see
    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers for full list.
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 3',
        'Topic :: Scientific/Engineering',
    ],
)



from handyscope.oscilloscopeChannel import OscilloscopeChannel
import math
import pytest
import time


def test_channel_cnt(osc):
    assert type(osc.channel_cnt) is int
    assert osc.channel_cnt > 0


def test_channels(osc):
    assert type(osc.channels) is tuple
    assert len(osc.channels) is osc.channel_cnt
    for channel in osc.channels:
        assert type(channel) is OscilloscopeChannel


def test__get_sample_cnts(osc):
    assert len(osc._get_sample_cnts()) == 2
    for element in osc._get_sample_cnts():
        assert type(element) is int
        assert element >= 0


def test_retrieve(osc):
    # Deactivate every channel
    for channel in osc.channels:
        channel.is_enabled = False
    with pytest.raises(ValueError):
        osc.retrieve()

    # Incrementally activate channels and call retrieve
    for idx, channel in enumerate(osc.channels):
        channel.is_enabled = True

        osc.start()
        while not osc.is_data_ready:
            time.sleep(0.05)
        data = osc.retrieve()

        assert type(data) is list
        assert len(data) == idx + 1
        for channel_data in data:
            assert type(channel_data) is list
            for sample in channel_data:
                assert type(sample) is float

    # Deactivate all except the last channel
    for channel in osc.channels[0:-1]:
        channel.is_enabled = False
    osc.channels[-1].is_enabled = True

    osc.start()
    while not osc.is_data_ready:
        time.sleep(0.05)
    data = osc.retrieve()

    assert type(data) is list
    assert len(data) == osc.channel_cnt
    # Only the last channel should contain actual samples
    for channel_data in data[0:-1]:
        assert channel_data is None
    assert type(data[-1]) is list
    for sample in data[-1]:
        assert type(sample) is float

    # Test parameter channel_nos: Get deactivated channel -> should raise a ValueError
    osc.channels[0].is_enabled = False
    osc.channels[1].is_enabled = True
    osc.start()
    while not osc.is_data_ready:
        time.sleep(0.05)
    with pytest.raises(ValueError):
        osc.retrieve(channel_nos=[1])

    # Test parameter channel_nos: Get activated channel -> should work
    osc.channels[0].is_enabled = False
    osc.channels[1].is_enabled = True
    osc.start()
    while not osc.is_data_ready:
        time.sleep(0.05)
    data = osc.retrieve(channel_nos=[2])
    assert type(data) is list
    assert len(data) is 2
    assert data[0] is None
    assert type(data[1]) is list
    for sample in data[1]:
        assert type(sample) is float


def test_retrieve_ch1(osc):
    # Enable available channels
    for channel in osc.channels:
        channel.is_enabled = True
    osc.start()
    while not osc.is_data_ready:
        time.sleep(0.05)
    data = osc.retrieve_ch1()
    assert type(data) is list
    assert len(data) == 1
    for idx, channel_data in enumerate(data):
        # Check, if channel is valid
        if idx < osc.channel_cnt:
            # Valid channel -> there must be valid data
            assert type(channel_data) is list
            for sample in channel_data:
                assert type(sample) is float
        else:
            # Invalid channel -> None
            assert channel_data is None

    # Check for multiple channels
    if osc.channel_cnt >= 2:
        # There is more than one channel -> we can disable the first without getting an error
        osc.channels[0].is_enabled = False
        osc.start()
        while not osc.is_data_ready:
            time.sleep(0.05)
        data = osc.retrieve_ch1()
        assert data == [None]


def test_retrieve_ch1_to_ch2(osc):
    # --- Try to retrieve all channels
    # Enable all available channels
    for channel in osc.channels:
        channel.is_enabled = True
    osc.start()
    while not osc.is_data_ready:
        time.sleep(0.05)
    data = osc.retrieve_ch1_to_ch2()
    assert type(data) is list
    assert len(data) == 2
    for idx, channel_data in enumerate(data):
        # Check, if channel is valid
        if idx < osc.channel_cnt:
            # Valid channel -> there must be valid data
            assert type(channel_data) is list
            for sample in channel_data:
                assert type(sample) is float
        else:
            # Invalid channel -> None
            assert channel_data is None

    # --- Try to retrieve all except the first channel
    # Check for multiple channels
    if osc.channel_cnt >= 2:
        # There is more than one channel -> we can disable the first without getting an error
        osc.channels[0].is_enabled = False
        osc.start()
        while not osc.is_data_ready:
            time.sleep(0.05)
        data = osc.retrieve_ch1_to_ch2()
        assert type(data) is list
        assert len(data) == 2
        # First channel was disabled
        assert data[0] is None
        # Second channel should contain data
        assert type(data[1]) is list
        for sample in data[1]:
            assert type(sample) is float


def test_retrieve_ch1_to_ch3(osc):
    # --- Try to retrieve all channels
    # Enable available channels
    for channel in osc.channels:
        channel.is_enabled = True
    osc.start()
    while not osc.is_data_ready:
        time.sleep(0.05)
    data = osc.retrieve_ch1_to_ch3()
    assert type(data) is list
    assert len(data) == 3
    for idx, channel_data in enumerate(data):
        # Check, if channel is valid
        if idx < osc.channel_cnt:
            # Valid channel -> there must be valid data
            assert type(channel_data) is list
            for sample in channel_data:
                assert type(sample) is float
        else:
            # Invalid channel -> None
            assert channel_data is None

    # --- Try to retrieve all except the first channel
    # Check for multiple channels
    if osc.channel_cnt >= 2:
        # There is more than one channel -> we can disable the first without getting an error
        osc.channels[0].is_enabled = False
        osc.start()
        while not osc.is_data_ready:
            time.sleep(0.05)
        data = osc.retrieve_ch1_to_ch3()

        assert type(data) is list
        assert len(data) == 3
        # First channel was disabled
        assert data[0] is None
        # Other channels should contain data, if they are valid ones
        for idx, channel_data in enumerate(data[1:]):
            # Check, if channel is valid
            if idx+1 < osc.channel_cnt:
                # Valid channel -> there must be valid data
                assert type(channel_data) is list
                for sample in channel_data:
                    assert type(sample) is float
            else:
                # Invalid channel -> None
                assert channel_data is None


def test_retrieve_ch1_to_ch4(osc):
    # --- Try to retrieve all channels
    # Enable available channels
    for channel in osc.channels:
        channel.is_enabled = True
    osc.start()
    while not osc.is_data_ready:
        time.sleep(0.05)
    data = osc.retrieve_ch1_to_ch4()
    assert type(data) is list
    assert len(data) == 4
    for idx, channel_data in enumerate(data):
        # Check, if channel is valid
        if idx < osc.channel_cnt:
            # Valid channel -> there must be valid data
            assert type(channel_data) is list
            for sample in channel_data:
                assert type(sample) is float
        else:
            # Invalid channel -> None
            assert channel_data is None

    # --- Try to retrieve all except the first channel
    # Check for multiple channels
    if osc.channel_cnt >= 2:
        # There is more than one channel -> we can disable the first without getting an error
        osc.channels[0].is_enabled = False
        osc.start()
        while not osc.is_data_ready:
            time.sleep(0.05)
        data = osc.retrieve_ch1_to_ch4()

        assert type(data) is list
        assert len(data) == 4
        # First channel was disabled
        assert data[0] is None
        # Other channels should contain data, if they are valid ones
        for idx, channel_data in enumerate(data[1:]):
            # Check, if channel is valid
            if idx+1 < osc.channel_cnt:
                # Valid channel -> there must be valid data
                assert type(channel_data) is list
                for sample in channel_data:
                    assert type(sample) is float
            else:
                # Invalid channel -> None
                assert channel_data is None


def test_valid_pre_sample_cnt(osc):
    assert type(osc.valid_pre_sample_cnt) is int
    assert osc.valid_pre_sample_cnt >= 0


def test_start(default_osc):
    # Starting a measurement should work without problems and return True
    assert default_osc.start() is True


def test_stop(default_osc):
    # Stopping a measurement right after start should return True
    default_osc.start()
    assert default_osc.stop() is True


def test_force_trig(default_osc):
    # Disable triggers
    for channel in default_osc.channels:
        channel.trig_is_enabled = False
    default_osc.trig_timeout = -1

    assert default_osc.is_force_trig is False

    # Start measurement and force trigger
    default_osc.start()
    default_osc.force_trig()

    while not default_osc.is_data_ready:
        time.sleep(0.01)
    assert default_osc.is_force_trig is True


def test_measure_modes_available(default_osc):
    assert type(default_osc.measure_modes_available) is tuple
    for mode in default_osc.measure_modes_available:
        assert mode in default_osc.MEASURE_MODES


def test_measure_mode(default_osc):
    # Test getter
    assert default_osc.measure_mode in default_osc.MEASURE_MODES

    # Test setter
    for mode in default_osc.measure_modes_available:
        default_osc.measure_mode = mode
        assert default_osc.measure_mode == mode


def test_is_running(default_osc):
    assert default_osc.is_running is False
    default_osc.start()
    assert default_osc.is_running is True


def test_is_triggered(default_osc):
    assert default_osc.is_triggered is False

    # Start measurement and force trigger
    default_osc.start()
    default_osc.force_trig()

    while not default_osc.is_data_ready:
        time.sleep(0.01)
    assert default_osc.is_triggered is True


def test_is_timeout_trig(default_osc):
    assert default_osc.is_timeout_trig is False
    default_osc.trig_timeout = 1e-6

    default_osc.start()

    while not default_osc.is_data_ready:
        time.sleep(0.01)
    assert default_osc.is_timeout_trig is True


def test_is_force_trig(default_osc):
    # Disable triggers
    for channel in default_osc.channels:
        channel.trig_is_enabled = False
    default_osc.trig_timeout = -1

    assert default_osc.is_force_trig is False

    # Start measurement and force trigger
    default_osc.start()
    default_osc.force_trig()

    while not default_osc.is_data_ready:
        time.sleep(0.01)
    assert default_osc.is_force_trig is True


def test_is_data_ready(default_osc):
    assert type(default_osc.is_data_ready) is bool

    default_osc.start()
    default_osc.force_trig()
    time.sleep(0.5)
    assert default_osc.is_data_ready is True


def test_is_data_overflow(default_osc):
    assert type(default_osc.is_data_overflow) is bool

    # Overflow can only occur in streaming mode
    default_osc.measure_mode = "stream"

    default_osc.start()

    # There should be no overflow in the beginning
    assert default_osc.is_data_overflow is False
    # No readout of data -> buffer should run full after some time
    time.sleep(1)
    assert default_osc.is_data_overflow is True


def test_resolutions_available(default_osc):
    assert type(default_osc.resolutions_available) is tuple
    for resolution in default_osc.resolutions_available:
        assert type(resolution) is int
        assert resolution > 0


def test_resolution(default_osc):
    # Test getter
    assert type(default_osc.resolution) is int
    assert default_osc.resolution > 0
    assert default_osc.resolution in default_osc.resolutions_available

    # Test setter
    for res in default_osc.resolutions_available:
        default_osc.resolution = res
        assert default_osc.resolution == res


def test_is_resolution_enhanced(default_osc):
    # Test for every available resolution
    for res in default_osc.resolutions_available:
        default_osc.resolution = res
        assert type(default_osc.is_resolution_enhanced) is bool


def test_auto_resolutions_available(default_osc):
    for res in default_osc.auto_resolutions_available:
        assert res in default_osc.AUTO_RESOLUTIONS


def test_auto_resolution(default_osc):
    # Test getter
    assert default_osc.auto_resolution in default_osc.AUTO_RESOLUTIONS

    # Test setter
    for res in default_osc.auto_resolutions_available:
        default_osc.auto_resolution = res
        assert default_osc.auto_resolution == res


def test_clock_sources_available(default_osc):
    for clk_src in default_osc.clock_sources_available:
        assert clk_src in default_osc.CLOCK_SOURCES


def test_clock_source(default_osc):
    # Test getter
    assert default_osc.clock_source in default_osc.CLOCK_SOURCES

    # Test setter
    for clk_src in default_osc.clock_sources_available:
        default_osc.clock_source = clk_src
        assert default_osc.clock_source == clk_src


def test_clock_outputs_available(default_osc):
    for clk_out in default_osc.clock_outputs_available:
        assert clk_out in default_osc.CLOCK_OUTPUTS


def test_clock_output(default_osc):
    # Test getter
    assert default_osc.clock_output in default_osc.CLOCK_OUTPUTS

    # Test setter
    for clk_out in default_osc.clock_outputs_available:
        default_osc.clock_output = clk_out
        assert default_osc.clock_output == clk_out


def test_sample_freq_max(default_osc):
    assert type(default_osc.sample_freq_max) is float
    assert default_osc.sample_freq_max > 0


def test_sample_freq(default_osc):
    # Test getter
    assert type(default_osc.sample_freq) is float
    assert default_osc.sample_freq > 0
    assert default_osc.sample_freq <= default_osc.sample_freq_max

    # Test setter
    freqs = [10e3, 100e3, default_osc.sample_freq_max]
    for freq in freqs:
        default_osc.sample_freq = freq
        assert default_osc.sample_freq == freq


def test_verify_sample_freq(default_osc):
    # Test type
    assert type(default_osc.verify_sample_freq(
        default_osc.sample_freq_max / 2)) is float
    # Test verify
    assert math.isclose(default_osc.verify_sample_freq(
        default_osc.sample_freq_max / 2),
        default_osc.sample_freq_max / 2, rel_tol=1e-6)
    # Test if the frequency clips
    assert default_osc.verify_sample_freq(
        default_osc.sample_freq_max + 1.0) == default_osc.sample_freq_max


def test_record_length_max(default_osc):
    assert type(default_osc.record_length_max) is int
    assert default_osc.record_length_max > 0


def test_record_length(default_osc):
    # Test getter
    assert type(default_osc.record_length) is int
    assert default_osc.record_length > 0
    assert default_osc.record_length <= default_osc.record_length_max

    # Test setter
    lengths = [1000, 10000, default_osc.record_length_max]
    for length in lengths:
        default_osc.record_length = length
        assert default_osc.record_length == length


def test_verify_record_length(default_osc):
    # Test type
    assert type(default_osc.verify_record_length(
        default_osc.record_length_max // 2)) is int
    # Test verify
    assert default_osc.verify_record_length(
        default_osc.record_length_max // 2) == default_osc.record_length_max // 2


def test_pre_sample_ratio(default_osc):
    # Test getter
    assert type(default_osc.pre_sample_ratio) is float
    assert default_osc.pre_sample_ratio >= 0
    assert default_osc.pre_sample_ratio <= 1

    # Test setter
    ratios = [0.0, 0.25, 0.5, 0.75, 1.0]
    for ratio in ratios:
        default_osc.pre_sample_ratio = ratio
        assert default_osc.pre_sample_ratio == ratio


def test_segment_cnt_max(default_osc):
    assert type(default_osc.segment_cnt_max) is int
    assert default_osc.segment_cnt_max >= 0


def test_segment_cnt(default_osc):
    # Test getter
    assert type(default_osc.segment_cnt) is int
    assert default_osc.segment_cnt > 0

    # Test setter
    cnts = [1, 2, default_osc.segment_cnt_max]
    for cnt in cnts:
        default_osc.segment_cnt = cnt
        assert default_osc.segment_cnt == cnt


def test_verify_segment_cnt(default_osc):
    # Test type
    assert type(default_osc.verify_segment_cnt(
        default_osc.segment_cnt_max // 2)) is int
    # Test verify
    assert default_osc.verify_segment_cnt(
        default_osc.segment_cnt_max // 2) == default_osc.segment_cnt_max // 2
    # Test if the segment cnt clips
    assert default_osc.verify_segment_cnt(
        default_osc.segment_cnt_max + 1) == default_osc.segment_cnt_max


def test_trig_timeout(default_osc):
    # Test getter
    assert type(default_osc.trig_timeout) is float
    assert (default_osc.trig_timeout >= 0) | (default_osc.trig_timeout == -1)

    # Test setter
    timeouts = [-1, 0, 1e-3, 0.5, 1]
    for timeout in timeouts:
        default_osc.trig_timeout = timeout
        assert default_osc.trig_timeout == timeout


def test_verify_trig_timeout(default_osc):
    # Test type
    assert type(default_osc.verify_trig_timeout(1.0)) is float
    # Test verify
    assert default_osc.verify_trig_timeout(1.0) == 1.0


def test_is_trig_delay_available(default_osc):
    assert type(default_osc.is_trig_delay_available) is bool
    assert default_osc.is_trig_delay_available is True


def test_trig_delay_max(default_osc):
    assert type(default_osc.trig_delay_max) is float
    assert default_osc.trig_delay_max >= 0


def test_trig_delay(default_osc):
    # Test getter
    assert type(default_osc.trig_delay) is float
    assert default_osc.trig_delay >= 0

    # Test setter
    delays = [0, 1e-3, default_osc.trig_delay_max]
    for delay in delays:
        default_osc.trig_delay = delay
        assert default_osc.trig_delay == delay


def test_verify_trig_delay(default_osc):
    # Test type
    assert type(default_osc.verify_trig_delay(
        default_osc.trig_delay_max / 2)) is float
    # Test verify
    assert math.isclose(
        default_osc.verify_trig_delay(default_osc.trig_delay_max / 2),
        default_osc.trig_delay_max / 2,
        rel_tol=1e-6)
    # Test if the segment cnt clips
    assert default_osc.verify_trig_delay(
        default_osc.trig_delay_max + 1.0) == default_osc.trig_delay_max


def test_is_trig_holdoff_available(default_osc):
    assert type(default_osc.is_trig_holdoff_available) is bool


def test_trig_holdoff_max(default_osc):
    assert type(default_osc.trig_holdoff_max) is int
    assert default_osc.trig_holdoff_max >= 0


def test_trig_holdoff(default_osc):
    # Test getter
    assert type(default_osc.trig_holdoff) is int
    assert default_osc.trig_holdoff >= 0

    # Test setter
    holdoffs = [0, 1, 100, default_osc.trig_holdoff_max, default_osc.TRIG_HOLDOFF_ALL_PRE_SAMPLES]
    for holdoff in holdoffs:
        default_osc.trig_holdoff = holdoff
        assert default_osc.trig_holdoff == holdoff


def test_is_trig_available(default_osc):
    assert type(default_osc.is_trig_available) is bool


def test_is_connection_test_available(default_osc):
    assert type(default_osc.is_connection_test_available) is bool


def test_start_connection_test(default_osc):
    if default_osc.is_connection_test_available:
        assert default_osc.start_connection_test() is True
        # Wait until connection test is completed before returning (there is no known way to stop the test)
        while not default_osc.is_connection_test_completed:
            time.sleep(0.01)
    else:
        with pytest.raises(OSError) as err:
            default_osc.start_connection_test()
            assert err.value.args[0] == "[-2]: NOT_SUPPORTED"


def test_is_connection_test_completed(default_osc):
    if default_osc.is_connection_test_available:
        assert type(default_osc.is_connection_test_completed) is bool
    else:
        with pytest.raises(OSError) as err:
            default_osc.is_connection_test_completed
            assert err.value.args[0] == "[-2]: NOT_SUPPORTED"


def test_connection_test_data(default_osc):
    if default_osc.is_connection_test_available:
        # Start the test
        default_osc.start_connection_test()
        # Wait, until it's finished
        while not default_osc.is_connection_test_completed:
            time.sleep(0.01)

        # Get the results
        results = default_osc.connection_test_data
        assert type(results) is tuple
        for idx, result in enumerate(results):
            if result == "undefined":
                assert default_osc.channels[idx].is_enabled is False
            else:
                assert default_osc.channels[idx].is_enabled is True
    else:
        with pytest.raises(OSError) as err:
            default_osc.connection_test_data
            assert err.value.args[0] == "[-2]: NOT_SUPPORTED"


def test_test_connection(default_osc):
    results = default_osc.test_connection()

    if default_osc.is_connection_test_available:
        assert type(results) is tuple
        for idx, result in enumerate(results):
            if result == "undefined":
                assert default_osc.channels[idx].is_enabled is False
            else:
                assert default_osc.channels[idx].is_enabled is True
    else:
        assert results is None


def test_measure(default_osc):
    data = default_osc.measure()
    assert type(data) is list
    # default_osc has only one enabled channel
    assert len(data) == 1
    assert type(data[0]) is list
    for sample in data[0]:
        assert type(sample) is float

    # Check for multiple channels
    if default_osc.channel_cnt >= 2:
        # There is more than one channel -> disable first, enable second
        default_osc.channels[0].is_enabled = False
        default_osc.channels[1].is_enabled = True

        data = default_osc.measure()

        assert type(data) is list
        assert len(data) == 2
        # First channel was disabled
        assert data[0] is None
        # Second channel should contain data
        assert type(data[1]) is list
        for sample in data[1]:
            assert type(sample) is float


def test_time_vector(default_osc):
    default_osc.pre_sample_ratio = 0
    assert len(default_osc.time_vector) == default_osc.record_length
    assert default_osc.time_vector[0] == 0
    assert default_osc.time_vector[1] == pytest.approx(1/default_osc.sample_freq)
    assert default_osc.time_vector[-1] == 1/default_osc.sample_freq*(default_osc.record_length-1)

    default_osc.pre_sample_ratio = 0.5
    # With correctly configured pre samples, it should work
    default_osc.trig_holdoff = default_osc.TRIG_HOLDOFF_ALL_PRE_SAMPLES
    assert len(default_osc.time_vector) == default_osc.record_length
    trig_idx = int(default_osc.record_length * default_osc.pre_sample_ratio)
    assert default_osc.time_vector[trig_idx] == 0
    assert default_osc.time_vector[trig_idx+1] == pytest.approx(1/default_osc.sample_freq)
    assert default_osc.time_vector[-1] == 1/default_osc.sample_freq*(default_osc.record_length-trig_idx-1)
    assert default_osc.time_vector[0] == -1/default_osc.sample_freq*(trig_idx)



from handyscope.oscilloscopeChannel import OscilloscopeChannel
import pytest


def test_bandwidths_available(osc):
    for channel in osc.channels:
        assert type(channel.bandwidths_available) is tuple


def test_bandwidth(osc):
    for channel in osc.channels:
        assert type(channel.bandwidth) is float
        assert channel.bandwidth >= 0


def test_is_safe_ground_available(osc):
    for channel in osc.channels:
        assert type(channel.is_safe_ground_available) is bool


def test_connector_type(osc):
    for channel in osc.channels:
        assert channel.connector_type in OscilloscopeChannel.CONNECTOR_TYPES


def test_is_differential(osc):
    for channel in osc.channels:
        assert type(channel.is_differential) is bool


def test_impedance(osc):
    for channel in osc.channels:
        assert type(channel.impedance) is float
        assert channel.impedance > 0


def test_couplings_available(osc):
    for channel in osc.channels:
        # must at least contain "unknown"
        assert len(channel.couplings_available) > 0

        for coupling in channel.couplings_available:
            assert coupling in OscilloscopeChannel.COUPLINGS


def test_coupling(osc):
    for channel in osc.channels:
        # test getter
        assert channel.coupling in OscilloscopeChannel.COUPLINGS

        # test setter
        for coupling in channel.couplings_available:
            channel.coupling = coupling
            assert channel.coupling is coupling


def test_is_enabled(osc):
    for channel in osc.channels:
        # test getter
        assert type(channel.is_enabled) is bool

        # test setter
        channel.is_enabled = True
        assert channel.is_enabled is True
        channel.is_enabled = False
        assert channel.is_enabled is False


def test_probe_gain(osc):
    for channel in osc.channels:
        # test getter
        assert type(channel.probe_gain) is float
        assert -1e6 <= channel.probe_gain <= 1e6
        assert channel.probe_gain != 0

        # test setter
        # valid gains
        for valid_gain in [-1e6, 1, 1e6]:
            channel.probe_gain = valid_gain
            assert channel.probe_gain == valid_gain

        # invalid gain (0)
        with pytest.raises(OSError) as err:
            channel.probe_gain = 0
        assert err.value.args[0] == "[-4]: INVALID_VALUE"

        # gains out of range
        for oor_gain in [-2e6, 2e6]:
            with pytest.warns(UserWarning) as record:
                channel.probe_gain = oor_gain
            # check that only one warning was raised
            assert len(record) == 1
            # check that the message matches
            assert record[0].message.args[0] == "[1]: VALUE_CLIPPED"


def test_probe_offset(osc):
    for channel in osc.channels:
        # test getter
        assert type(channel.probe_offset) is float
        assert -1e6 <= channel.probe_offset <= 1e6

        # test setter
        # valid offsets
        for valid_offset in [-1e6, 0, 1, 1e6]:
            channel.probe_offset = valid_offset
            assert channel.probe_offset == valid_offset

        # offsets out of range
        for oor_offset in [-2e6, 2e6]:
            with pytest.warns(UserWarning) as record:
                channel.probe_offset = oor_offset
            # check that only one warning was raised
            assert len(record) == 1
            # check that the message matches
            assert record[0].message.args[0] == "[1]: VALUE_CLIPPED"


def test_is_auto_range(osc):
    for channel in osc.channels:
        # test getter
        assert type(channel.is_auto_range) is bool

        # test setter
        channel.is_auto_range = True
        assert channel.is_auto_range is True
        channel.is_auto_range = False
        assert channel.is_auto_range is False


def test_ranges_available(osc):
    for channel in osc.channels:
        assert type(channel.ranges_available) is tuple
        for element in channel.ranges_available:
            assert type(element) is float
            assert element > 0


def test_range(osc):
    for channel in osc.channels:
        # Test getter
        assert type(channel.range) is float
        assert channel.range > 0

        # Test setter
        for range_available in channel.ranges_available:
            channel.range = range_available
            assert channel.range == range_available


def test_is_trig_enabled(osc):
    for channel in osc.channels:
        # test getter
        assert type(channel.is_trig_enabled) is bool

        # test setter
        channel.is_trig_enabled = True
        assert channel.is_trig_enabled is True
        channel.is_trig_enabled = False
        assert channel.is_trig_enabled is False


def test_trig_kinds_available(osc):
    for channel in osc.channels:
        for trig_kind in channel.trig_kinds_available:
            assert trig_kind in channel.TRIGGER_KINDS


def test_trig_kind(osc):
    for channel in osc.channels:
        # Test getter
        # If there are multiple kinds available, the chosen kind can be retrieved.
        if channel.trig_kinds_available != ("unknown", ):
            assert channel.trig_kind in channel.TRIGGER_KINDS
        # Else (no known kinds), accessing the chosen kind raises an OSError.
        else:
            with pytest.raises(OSError) as err:
                channel.trig_kind
            assert err.value.args[0] == "[-2]: NOT_SUPPORTED"

        # Test setter
        # If there are multiple kinds available, the chosen kind can be set.
        if channel.trig_kinds_available != ("unknown", ):
            # Test every available kind
            for kind in channel.trig_kinds_available:
                # Set the value
                channel.trig_kind = kind
                # Read it back & compare
                assert channel.trig_kind == kind
        # Else (no known kinds), setting the chosen kind raises an OSError.
        else:
            with pytest.raises(OSError) as err:
                channel.trig_kind = "unknown"
            assert err.value.args[0] == "[-2]: NOT_SUPPORTED"


def test_trig_lvl_cnt(osc):
    for channel in osc.channels:
        # Alter trig_kind, because trig_lvl_cnt is influenced by them
        for trig_kind in channel.trig_kinds_available:
            channel.trig_kind = trig_kind
            assert type(channel.trig_lvl_cnt) is int
            assert channel.trig_lvl_cnt >= 0


def test_trig_lvl(osc):
    for channel in osc.channels:
        # Alter trig_kind, because trig_lvl is influenced by them
        for trig_kind in channel.trig_kinds_available:
            channel.trig_kind = trig_kind

            # Test getter
            assert len(channel.trig_lvl) is channel.trig_lvl_cnt
            for element in channel.trig_lvl:
                assert type(element) is float

            # Test setter
            # Assuming the default trig_lvl_mode "relative", values range from 0 to 1
            channel.trig_lvl = tuple(0.0 for _ in range(channel.trig_lvl_cnt))
            assert channel.trig_lvl == tuple(0.0 for _ in range(channel.trig_lvl_cnt))
            channel.trig_lvl = tuple(0.5 for _ in range(channel.trig_lvl_cnt))
            assert channel.trig_lvl == tuple(0.5 for _ in range(channel.trig_lvl_cnt))
            channel.trig_lvl = tuple(1.0 for _ in range(channel.trig_lvl_cnt))
            assert channel.trig_lvl == tuple(1.0 for _ in range(channel.trig_lvl_cnt))


def test_trig_lvl_modes_available(osc):
    for channel in osc.channels:
        for mode in channel.trig_lvl_modes_available:
            assert mode in channel.TRIGGER_LVL_MODES


def test_trig_lvl_mode(osc):
    for channel in osc.channels:
        # Test getter
        # If there are multiple modes available, the mode can be retrieved.
        if channel.trig_lvl_modes_available != ("unknown", ):
            assert channel.trig_lvl_mode in channel.TRIGGER_LVL_MODES
        # Else (no known modes), accessing the chosen kind raises an OSError.
        else:
            with pytest.raises(OSError) as err:
                channel.trig_lvl_mode
            assert err.value.args[0] == "[-2]: NOT_SUPPORTED"

        # Test setter
        # If there are multiple modes available, the mode can be set.
        if channel.trig_lvl_modes_available != ("unknown", ):
            # Test every available mode
            for mode in channel.trig_lvl_modes_available:
                # Set the value
                channel.trig_lvl_mode = mode
                # Read it back & compare
                assert channel.trig_lvl_mode == mode
        # Else (no known modes), setting the mode raises an OSError.
        else:
            with pytest.raises(OSError) as err:
                channel.trig_lvl_mode = "unknown"
            assert err.value.args[0] == "[-2]: NOT_SUPPORTED"


def test_trig_hysteresis_cnt(osc):
    for channel in osc.channels:
        # Alter trig_kind, because trig_hysteresis_cnt is influenced by them
        for trig_kind in channel.trig_kinds_available:
            channel.trig_kind = trig_kind
            assert type(channel.trig_hysteresis_cnt) is int
            assert channel.trig_hysteresis_cnt >= 0
            
            
def test_trig_hysteresis(osc):
    for channel in osc.channels:
        # Alter trig_kind, because trig_hysteresis is influenced by them
        for trig_kind in channel.trig_kinds_available:
            channel.trig_kind = trig_kind

            # Test getter
            assert len(channel.trig_hysteresis) is channel.trig_hysteresis_cnt
            for element in channel.trig_hysteresis:
                assert type(element) is float

            # Test setter
            channel.trig_hysteresis = tuple(0.0 for _ in range(channel.trig_hysteresis_cnt))
            assert channel.trig_hysteresis == tuple(0.0 for _ in range(channel.trig_hysteresis_cnt))
            channel.trig_hysteresis = tuple(0.5 for _ in range(channel.trig_hysteresis_cnt))
            assert channel.trig_hysteresis == tuple(0.5 for _ in range(channel.trig_hysteresis_cnt))
            channel.trig_hysteresis = tuple(1.0 for _ in range(channel.trig_hysteresis_cnt))
            assert channel.trig_hysteresis == tuple(1.0 for _ in range(channel.trig_hysteresis_cnt))


def test_trig_conditions_available(osc):
    for channel in osc.channels:
        # Alter trig_kind, because trig_conditions_available is influenced by them
        for trig_kind in channel.trig_kinds_available:
            channel.trig_kind = trig_kind

            for trig_condition in channel.trig_conditions_available:
                assert trig_condition in channel.TRIGGER_CONDITIONS


def test_trig_condition(osc):
    for channel in osc.channels:
        # Alter trig_kind, because trig_condition is influenced by them
        for trig_kind in channel.trig_kinds_available:
            channel.trig_kind = trig_kind

            # Test getter
            # If there are multiple conditions available, the chosen condition can be retreived.
            if channel.trig_conditions_available != ("unknown", ):
                assert channel.trig_condition in channel.TRIGGER_CONDITIONS
            # Else (no known conditions), accessing the chosen condition raises an OSError.
            else:
                with pytest.raises(OSError) as err:
                    channel.trig_condition
                assert err.value.args[0] == "[-2]: NOT_SUPPORTED"

            # Test setter
            # If there are multiple conditions available, the condition can be set.
            if channel.trig_conditions_available != ("unknown", ):
                for condition in channel.trig_conditions_available:
                    channel.trig_condition = condition
                    assert channel.trig_condition is condition
            # Else (no known conditions), setting the chosen condition raises an OSError.
            else:
                with pytest.raises(OSError) as err:
                    channel.trig_condition = "unknown"
                assert err.value.args[0] == "[-2]: NOT_SUPPORTED"


def test_trig_time_cnt(osc):
    for channel in osc.channels:
        # Alter trig_kind, because trig_time_cnt is influenced by it
        for trig_kind in channel.trig_kinds_available:
            channel.trig_kind = trig_kind
            # If possible, alter trig_condition, because trig_time_cnt is influenced by it
            if channel.trig_conditions_available != ("unknown", ):
                for trig_condition in channel.trig_conditions_available:
                    channel.trig_condition = trig_condition

                    assert type(channel.trig_time_cnt) is int
                    assert channel.trig_time_cnt >= 0
            # If not possible, just do the check
            else:
                assert type(channel.trig_time_cnt) is int
                assert channel.trig_time_cnt >= 0


def test_trig_time(osc):
    for channel in osc.channels:
        # Alter trig_kind, because trig_time_cnt is influenced by it
        for trig_kind in channel.trig_kinds_available:
            channel.trig_kind = trig_kind
            # If possible, alter trig_condition, because trig_time_cnt is influenced by it
            if channel.trig_conditions_available != ("unknown", ):
                for trig_condition in channel.trig_conditions_available:
                    channel.trig_condition = trig_condition

                    # Test getter
                    assert type(channel.trig_time) is tuple
                    assert len(channel.trig_time) is channel.trig_time_cnt
                    for element in channel.trig_time:
                        assert type(element) is float

                    # Test setter
                    channel.trig_time = tuple(0.1 for _ in range(channel.trig_time_cnt))
                    assert channel.trig_time == tuple(0.1 for _ in range(channel.trig_time_cnt))
                    channel.trig_time = tuple(0.001 for _ in range(channel.trig_time_cnt))
                    assert channel.trig_time == tuple(0.001 for _ in range(channel.trig_time_cnt))

            # If not possible, just do the check
            else:
                # Test getter
                assert type(channel.trig_time) is tuple
                assert len(channel.trig_time) is channel.trig_time_cnt
                for element in channel.trig_time:
                    assert type(element) is float

                # Test setter
                channel.trig_time = tuple(0.1 for _ in range(channel.trig_time_cnt))
                assert channel.trig_time == tuple(0.1 for _ in range(channel.trig_time_cnt))
                channel.trig_time = tuple(0.001 for _ in range(channel.trig_time_cnt))
                assert channel.trig_time == tuple(0.001 for _ in range(channel.trig_time_cnt))


def test_is_trig_available(osc):
    for channel in osc.channels:
        assert type(channel.is_trig_available) is bool


def test_is_available(osc):
    for channel in osc.channels:
        assert type(channel.is_available) is bool


def test_is_connection_test_available(osc):
    for channel in osc.channels:
        assert type(channel.is_connection_test_available) is bool


def test_data_range(osc):
    for channel in osc.channels:
        assert type(channel.data_range) is tuple
        assert len(channel.data_range) is 2
        for element in channel.data_range:
            assert type(element) is float


def test_data_value_min(osc):
    for channel in osc.channels:
        assert type(channel.data_value_min) is float


def test_data_value_max(osc):
    for channel in osc.channels:
        assert type(channel.data_value_max) is float


def test_raw_data_type(osc):
    for channel in osc.channels:
        assert type(channel.raw_data_type) is str
        assert channel.raw_data_type in channel.RAW_DATA_TYPES.keys()


def test_raw_value_min(osc):
    for channel in osc.channels:
        assert type(channel.raw_value_min) is int
        assert channel.raw_value_min >= 0


def test_raw_value_max(osc):
    for channel in osc.channels:
        assert type(channel.raw_value_max) is int
        assert channel.raw_value_max >= 0
        assert channel.raw_value_max >= channel.raw_value_min


def test_raw_value_zero(osc):
    for channel in osc.channels:
        assert type(channel.raw_value_zero) is int
        assert channel.raw_value_zero >= channel.raw_value_min
        assert channel.raw_value_zero <= channel.raw_value_max


def test_raw_data_range(osc):
    for channel in osc.channels:
        assert type(channel.raw_data_range) is tuple
        assert channel.raw_data_range[0] == channel.raw_value_min
        assert channel.raw_data_range[1] == channel.raw_value_zero
        assert channel.raw_data_range[2] == channel.raw_value_max



from handyscope.library import libtiepie

from datetime import date
import ctypes


class DeviceList:
    """This class provides access to the device list maintained by libtiepie.

    Attributes:
        ID_KINDS (dict): dict which maps readable representations of id kinds
                         to their int version
        PRODUCT_IDS (dict): dict which maps readable representations of product
                            ids to their int version
        DEVICE_TYPES (dict): dict which maps readable representations of
                             device types to their int version
    """

    ID_KINDS = {"product id":    1,
                "index":         2,
                "serial number": 4}

    PRODUCT_IDS = {"none":       0,
                   "combined":   2,
                   "HS3":       13,
                   "HS4":       15,
                   "HP3":       18,
                   "HS4D":      20,
                   "HS5":       22,
                   "HS6":       24,
                   "HS6D":      25}

    DEVICE_TYPES = {"Osc": 1,
                    "Gen": 2,
                    "I2C": 4}

    def __init__(self):
        """Constructor for class DeviceList."""
        # Fill the device list
        libtiepie.LstUpdate()

    @property
    def device_cnt(self):
        """Get device count.

        Returns:
            int: device count
        """
        return libtiepie.LstGetCount()

    def can_open_device(self, instr_id, id_kind="index", device_type="Osc"):
        """Check whether a device can be opened.

        Args:
            instr_id (int or str): Device list index, product ID
                                   (listed in dict PRODUCT_IDS) or
                                   serial number.
            id_kind (str): The kind of the given instr_id
                           (listed in dict ID_KINDS), defaults to device
                           list index.
            device_type (str): The type of the device
                               (listed in dict DEVICE_TYPES),
                               defaults to oscilloscope.

        Returns:
            bool: True, if the device can be opened.
        """
        # translate id kind & device type str to int
        id_kind_int = self.ID_KINDS[id_kind]
        device_type_int = self.DEVICE_TYPES[device_type]

        # Translate instr_id to int, if it is a product id str
        if id_kind == "product id":
            instr_id_int = self.PRODUCT_IDS[instr_id]
        else:
            instr_id_int = instr_id

        return libtiepie.LstDevCanOpen(id_kind_int,
                                       instr_id_int, device_type_int) == 1

    def get_vendor_id(self, instr_id, id_kind="index"):
        """Get the vendor id of the device.

        Args:
            instr_id (int): Device list index, product ID
                            (listed in dict PRODUCT_IDS) or serial number.
            id_kind (str): the kind of the given instr_id
                           (listed in dict ID_KINDS), defaults to device list
                           index.

        Returns:
            int: vendor id
        """
        # translate id kind str to int
        id_kind_int = self.ID_KINDS[id_kind]

        # Translate instr_id to int, if it is a product id str
        if id_kind == "product id":
            instr_id_int = self.PRODUCT_IDS[instr_id]
        else:
            instr_id_int = instr_id
        # get the vendor id
        vendor_id = libtiepie.LstDevGetVendorId(id_kind_int, instr_id_int)

        return vendor_id

    def get_product_id(self, instr_id, id_kind="index"):
        """Get the product id of the device.

        Args:
            instr_id (int): Device list index, product ID
                            (listed in dict PRODUCT_IDS) or serial number.
            id_kind (str): the kind of the given instr_id
                           (listed in dict ID_KINDS), defaults to device list
                           index.

        Returns:
            int: product id
        """
        # translate id kind str to int
        id_kind_int = self.ID_KINDS[id_kind]

        # Translate instr_id to int, if it is a product id str
        if id_kind == "product id":
            instr_id_int = self.PRODUCT_IDS[instr_id]
        else:
            instr_id_int = instr_id
        # get the product id
        product_id = libtiepie.LstDevGetProductId(id_kind_int, instr_id_int)

        return product_id

    def get_driver_version(self, instr_id, id_kind="index"):
        """Get the driver version of the device.

        Args:
            instr_id (int): Device list index, product ID
                            (listed in dict PRODUCT_IDS) or serial number.
            id_kind (str): the kind of the given instr_id
                           (listed in dict ID_KINDS), defaults to device list
                           index.

        Returns:
            str: driver version
        """
        # translate id kind str to int
        id_kind_int = self.ID_KINDS[id_kind]

        # Translate instr_id to int, if it is a product id str
        if id_kind == "product id":
            instr_id_int = self.PRODUCT_IDS[instr_id]
        else:
            instr_id_int = instr_id
        # get the driver version
        driver_version = libtiepie.LstDevGetDriverVersion(id_kind_int,
                                                          instr_id_int)

        return version_to_str(driver_version)

    def get_recommended_driver_version(self, instr_id, id_kind="index"):
        """Get the recommended driver version of the device.

        Not tested.
        
        Args:
            instr_id (int): Device list index, product ID
                            (listed in dict PRODUCT_IDS) or serial number.
            id_kind (str): the kind of the given instr_id
                           (listed in dict ID_KINDS), defaults to device list
                           index.

        Returns:
            str: recommended driver version
        """
        # translate id kind str to int
        id_kind_int = self.ID_KINDS[id_kind]

        # Translate instr_id to int, if it is a product id str
        if id_kind == "product id":
            instr_id_int = self.PRODUCT_IDS[instr_id]
        else:
            instr_id_int = instr_id
        # get the driver version
        driver_version = libtiepie.LstDevGetRecommendedDriverVersion(
            id_kind_int, instr_id_int)

        return version_to_str(driver_version)

    def get_firmware_version(self, instr_id, id_kind="index"):
        """Get the firmware version of the device.

        Args:
            instr_id (int): Device list index, product ID
                            (listed in dict PRODUCT_IDS) or serial number.
            id_kind (str): the kind of the given instr_id
                           (listed in dict ID_KINDS), defaults to device list
                           index.

        Returns:
            str: firmware version
        """
        # translate id kind str to int
        id_kind_int = self.ID_KINDS[id_kind]

        # Translate instr_id to int, if it is a product id str
        if id_kind == "product id":
            instr_id_int = self.PRODUCT_IDS[instr_id]
        else:
            instr_id_int = instr_id
        # get the firmware version
        firmware_version = libtiepie.LstDevGetFirmwareVersion(id_kind_int,
                                                              instr_id_int)

        return version_to_str(firmware_version)

    def get_recommended_firmware_version(self, instr_id, id_kind="index"):
        """Get the recommended firmware version of the device.
        
        Not tested.

        Args:
            instr_id (int): Device list index, product ID
                            (listed in dict PRODUCT_IDS) or serial number.
            id_kind (str): the kind of the given instr_id
                           (listed in dict ID_KINDS), defaults to device list
                           index.

        Returns:
            str: recommended firmware version
        """
        # translate id kind str to int
        id_kind_int = self.ID_KINDS[id_kind]

        # Translate instr_id to int, if it is a product id str
        if id_kind == "product id":
            instr_id_int = self.PRODUCT_IDS[instr_id]
        else:
            instr_id_int = instr_id
        # get the firmware version
        firmware_version = libtiepie.LstDevGetRecommendedFirmwareVersion(
            id_kind_int, instr_id_int)

        return version_to_str(firmware_version)

    def get_calibration_date(self, instr_id, id_kind="index"):
        """Get the calibration date of the device.

        Args:
            instr_id (int): Device list index, product ID
                            (listed in dict PRODUCT_IDS) or serial number.
            id_kind (str): the kind of the given instr_id
                           (listed in dict ID_KINDS), defaults to device list
                           index.

        Returns:
            :py:class:`datetime.date`: calibration date
        """
        # translate id kind str to int
        id_kind_int = self.ID_KINDS[id_kind]

        # Translate instr_id to int, if it is a product id str
        if id_kind == "product id":
            instr_id_int = self.PRODUCT_IDS[instr_id]
        else:
            instr_id_int = instr_id

        # get the calibration date
        raw_date = libtiepie.LstDevGetCalibrationDate(id_kind_int, instr_id_int)

        split_date = date(raw_date >> 16, (raw_date >> 8) & 0xff,
                          raw_date & 0xff)
        return split_date

    def get_device_name(self, instr_id, id_kind="index"):
        """Get the full name of the device.

        Args:
            instr_id (int): Device list index, product ID
                            (listed in dict PRODUCT_IDS) or serial number.
            id_kind (str): the kind of the given instr_id
                           (listed in dict ID_KINDS), defaults to device list
                           index.

        Returns:
            str: full device name
        """
        # translate id kind str to int
        id_kind_int = self.ID_KINDS[id_kind]

        # get length of device name string
        str_len = libtiepie.LstDevGetName(id_kind_int, instr_id, None, 0)

        # initialize mutable string buffer
        str_buffer = ctypes.create_string_buffer(str_len)

        # write the actual device name to the buffer
        libtiepie.LstDevGetName(id_kind_int, instr_id, str_buffer, str_len)

        # convert to a normal python string
        dev_name = str_buffer.value.decode('utf-8')

        return dev_name

    def get_device_name_short(self, instr_id, id_kind="index"):
        """Get the short name of the device.

        Args:
            instr_id (int): Device list index, product ID
                            (listed in dict PRODUCT_IDS) or serial number.
            id_kind (str): the kind of the given instr_id
                           (listed in dict ID_KINDS), defaults to device list
                           index.

        Returns:
            str: short device name
        """
        # translate id kind str to int
        id_kind_int = self.ID_KINDS[id_kind]

        # get length of device name string
        str_len = libtiepie.LstDevGetNameShort(id_kind_int, instr_id, None, 0)

        # initialize mutable string buffer
        str_buffer = ctypes.create_string_buffer(str_len)

        # write the actual device name to the buffer
        libtiepie.LstDevGetNameShort(id_kind_int, instr_id, str_buffer, str_len)

        # convert to a normal python string
        dev_name = str_buffer.value.decode('utf-8')

        return dev_name

    def get_device_name_shortest(self, instr_id, id_kind="index"):
        """Get the shortest name of the device.

        Args:
            instr_id (int): Device list index, product ID
                            (listed in dict PRODUCT_IDS) or serial number.
            id_kind (str): the kind of the given instr_id
                           (listed in dict ID_KINDS), defaults to device list
                           index.

        Returns:
            str: shortest device name
        """
        # translate id kind str to int
        id_kind_int = self.ID_KINDS[id_kind]

        # get length of device name string
        str_len = libtiepie.LstDevGetNameShortest(id_kind_int, instr_id,
                                                  None, 0)

        # initialize mutable string buffer
        str_buffer = ctypes.create_string_buffer(str_len)

        # write the actual device name to the buffer
        libtiepie.LstDevGetNameShortest(id_kind_int, instr_id, str_buffer,
                                        str_len)

        # convert to a normal python string
        dev_name = str_buffer.value.decode('utf-8')

        return dev_name

    def get_device_serial_no(self, instr_id, id_kind="index"):
        """Get the serial number of the device.

        Args:
            instr_id (int): Device list index, product ID
                            (listed in dict PRODUCT_IDS) or serial number.
            id_kind (str): the kind of the given instr_id
                           (listed in dict ID_KINDS), defaults to device list
                           index.

        Returns:
            int: serial number
        """
        # translate id kind str to int
        id_kind_int = self.ID_KINDS[id_kind]

        # get the serial number
        serial_no = libtiepie.LstDevGetSerialNumber(id_kind_int, instr_id)

        return serial_no

    def get_contained_serial_no(self, instr_id, id_kind="index"):
        """Get serial numbers of devices in a contained device.

        Args:
            instr_id (int): Device list index, product ID
                            (listed in dict PRODUCT_IDS) or serial number.
            id_kind (str): the kind of the given instr_id
                           (listed in dict ID_KINDS), defaults to device list
                           index.

        Returns:
            tuple: serial numbers of the devices
        """
        # translate id kind str to int
        id_kind_int = self.ID_KINDS[id_kind]

        # get length of list
        serial_len = libtiepie.LstDevGetContainedSerialNumbers(id_kind_int,
                                                               instr_id,
                                                               None, 0)
        # initialize uint32 array
        serial_numbers = (ctypes.c_uint32 * serial_len)()

        # write the actual data to the array
        libtiepie.LstDevGetContainedSerialNumbers(id_kind_int, instr_id,
                                                  serial_numbers, serial_len)

        # convert to a normal python list
        serial_numbers = list(serial_numbers)

        return tuple(serial_numbers)

    def get_ipv4_address(self, instr_id, id_kind="index"):
        """Get the IP address of an instrument.

        Args:
            instr_id (int): Device list index, product ID
                            (listed in dict PRODUCT_IDS) or serial number.
            id_kind (str): the kind of the given instr_id
                           (listed in dict ID_KINDS), defaults to device list
                           index.

        Returns:
            int: IP address
        """
        id_kind_int = self.ID_KINDS[id_kind]

        # Translate instr_id to int, if it is a product id str
        if id_kind == "product id":
            instr_id_int = self.PRODUCT_IDS[instr_id]
        else:
            instr_id_int = instr_id

        ip_v4_addr = libtiepie.LstDevGetIPv4Address(id_kind_int, instr_id_int)

        return ip_v4_addr

    def get_ip_port(self, instr_id, id_kind="index"):
        """Get the IP Port number of the instrument.

        Args:
            instr_id (int): Device list index, product ID
                            (listed in dict PRODUCT_IDS) or serial number.
            id_kind (str): the kind of the given instr_id
                           (listed in dict ID_KINDS), defaults to device list
                           index.

        Returns:
            int: IP port
        """
        id_kind_int = self.ID_KINDS[id_kind]

        # Translate instr_id to int, if it is a product id str
        if id_kind == "product id":
            instr_id_int = self.PRODUCT_IDS[instr_id]
        else:
            instr_id_int = instr_id

        ip_port = libtiepie.LstDevGetIPPort(id_kind_int, instr_id_int)

        return ip_port

    def is_connected_to_server(self, instr_id, id_kind="index"):
        """Get whether the device is connected to a server.

        Args:
            instr_id (int): Device list index, product ID
                            (listed in dict PRODUCT_IDS) or serial number.
            id_kind (str): the kind of the given instr_id
                           (listed in dict ID_KINDS), defaults to device list
                           index.

        Returns:
            bool: True, if the device is connected to a server.
        """
        id_kind_int = self.ID_KINDS[id_kind]

        # Translate instr_id to int, if it is a product id str
        if id_kind == "product id":
            instr_id_int = self.PRODUCT_IDS[instr_id]
        else:
            instr_id_int = instr_id

        return libtiepie.LstDevHasServer(id_kind_int, instr_id_int) == 1

    def get_device_types(self, instr_id, id_kind="index"):
        """Get the device types of an instrument.

        Args:
            instr_id (int): Device list index, product ID
                            (listed in dict PRODUCT_IDS) or serial number.
            id_kind (str): the kind of the given instr_id
                           (listed in dict ID_KINDS), defaults to device list
                           index.

        Returns:
            dict: key: type (as listed in DEVICE_TYPES), value: True/False.
        """
        # translate id kind str to int
        id_kind_int = self.ID_KINDS[id_kind]

        # get the device types
        dev_types = libtiepie.LstDevGetTypes(id_kind_int, instr_id)

        # check for every possible type and store in dict
        type_dict = {"Osc": False,
                     "Gen": False,
                     "I2C": False}
        for key in self.DEVICE_TYPES:
            if dev_types & self.DEVICE_TYPES[key]:
                type_dict[key] = True

        return type_dict

    def get_overview(self):
        """Return a list with information on all connected instruments (index,
        name, serial number and device types).

        Returns:
            tuple: Tuple of dicts with info on connected instruments.
        """
        instr_list = []
        for idx in range(self.device_cnt):
            instr_info = {"Index":      idx,
                          "Name":       self.get_device_name_short(idx),
                          "SerNo":      self.get_device_serial_no(idx),
                          "DevTypes":   self.get_device_types(idx)}
            instr_list.append(instr_info)

        return tuple(instr_list)

    def get_overview_str(self):
        """Return a string with information on all connected instruments
        (index, name, serial number and device types).

        Returns:
            str: Information on connected instruments.
        """
        instr_str = ""
        for instr_info in self.get_overview():
            instr_str += "%d:\t%s\t%d\t%s\n" % (instr_info["Index"],
                                                instr_info["Name"],
                                                instr_info["SerNo"],
                                                instr_info["DevTypes"])

        return instr_str

    def __str__(self):
        """Return a (more or less) human-readable representation of the
        instruments in the device list.

        Returns:
            str: Instruments in the device list and info on them.
        """
        return self.get_overview_str()

    def open_device(self, instr_id, id_kind="index", device_type="Osc"):
        """Open a device (of device_type) of an instrument (with given id) and
        return the device handle.

        Args:
            instr_id (int or str): Device list index, product ID (listed
                                   in dict PRODUCT_IDS) or serial number.
            id_kind (str): the kind of the given instr_id (listed in
                           dict ID_KINDS), defaults to device list index.
            device_type (str): the type of the device (listed in dict
                               DEVICE_TYPES), defaults to oscilloscope.

        Returns:
            handle (:py:class:`ctypes.c_uint32`): device handle
        """
        # translate id kind & device type str to int
        id_kind_int = self.ID_KINDS[id_kind]
        device_type_int = self.DEVICE_TYPES[device_type]

        # Translate instr_id to int, if it is a product id str
        if id_kind == "product id":
            instr_id_int = self.PRODUCT_IDS[instr_id]
        else:
            instr_id_int = instr_id

        return libtiepie.LstOpenDevice(id_kind_int, instr_id_int,
                                       device_type_int)

    def get_channel_count_cb(self, instr_id, contained_serial_no,
                             id_kind="index"):
        """Get the amount of channels of a contained oscilloscope in a combined
        instrument.

        Args:
            instr_id (int or str): Device list index, product ID (listed
                                   in dict PRODUCT_IDS) or serial number.
            id_kind (str): The kind of the given instr_id (listed in dict
                           ID_KINDS), defaults to device list index.
            contained_serial_no (int): Serial number of the desired device
                                       contained in the combined device.

        Returns:
            int: Channel count of the device
        """
        # translate id kind & device type str to int
        id_kind_int = self.ID_KINDS[id_kind]
        # Translate instr_id to int, if it is a product id str
        if id_kind == "product id":
            instr_id_int = self.PRODUCT_IDS[instr_id]
        else:
            instr_id_int = instr_id

        return libtiepie.LstCbScpGetChannelCount(id_kind_int, instr_id_int,
                                                 contained_serial_no)

    def get_product_id_cb(self, instr_id, contained_serial_no,
                          id_kind="index"):
        """Get the product id of a contained device in a combined instrument.

        Args:
            instr_id (int or str): Device list index, product ID (listed
                                   in dict PRODUCT_IDS) or serial number.
            id_kind (str): The kind of the given instr_id (listed in dict
                           ID_KINDS), defaults to device list index.
            contained_serial_no (int): Serial number of the desired device
                                       contained in the combined device.

        Returns: Product id of the device.
        """
        # translate id kind & device type str to int
        id_kind_int = self.ID_KINDS[id_kind]
        # Translate instr_id to int, if it is a product id str
        if id_kind == "product id":
            instr_id_int = self.PRODUCT_IDS[instr_id]
        else:
            instr_id_int = instr_id

        return libtiepie.LstCbDevGetProductId(id_kind_int, instr_id_int,
                                              contained_serial_no)

    def get_vendor_id_cb(self, instr_id, contained_serial_no,
                         id_kind="index"):
        """Get the vendor_id of a contained device in a combined instrument.

        Args:
            instr_id (int or str): Device list index, product ID (listed
                                   in dict PRODUCT_IDS) or serial number.
            id_kind (str): The kind of the given instr_id (listed in dict
                           ID_KINDS), defaults to device list index.
            contained_serial_no (int): Serial number of the desired device
                                       contained in the combined device.

        Returns:
            int: vendor id of the device.
        """
        # translate id kind & device type str to int
        id_kind_int = self.ID_KINDS[id_kind]
        # Translate instr_id to int, if it is a product id str
        if id_kind == "product id":
            instr_id_int = self.PRODUCT_IDS[instr_id]
        else:
            instr_id_int = instr_id

        return libtiepie.LstCbDevGetVendorId(id_kind_int, instr_id_int,
                                             contained_serial_no)

    def get_device_name_cb(self, instr_id, contained_serial_no,
                           id_kind="index"):
        """Get the full name of a contained device in a combined instrument.

        Args:
            instr_id (int or str): Device list index, product ID (listed
                                   in dict PRODUCT_IDS) or serial number.
            id_kind (str): The kind of the given instr_id (listed in dict
                           ID_KINDS), defaults to device list index.
            contained_serial_no (int): Serial number of the desired device
                                       contained in the combined device.

        Returns:
            str: full device name of the device
        """
        # translate id kind str to int
        id_kind_int = self.ID_KINDS[id_kind]

        # get length of device name string
        str_len = libtiepie.LstCbDevGetName(id_kind_int, instr_id,
                                            contained_serial_no, None, 0)

        # initialize mutable string buffer
        str_buffer = ctypes.create_string_buffer(str_len)

        # write the actual device name to the buffer
        libtiepie.LstCbDevGetName(id_kind_int, instr_id, contained_serial_no,
                                  str_buffer, str_len)

        # convert to a normal python string
        dev_name = str_buffer.value.decode('utf-8')

        return dev_name

    def get_device_name_short_cb(self, instr_id, contained_serial_no,
                                 id_kind="index"):
        """Get the short name of a contained device in a combined instrument.

        Args:
            instr_id (int or str): Device list index, product ID (listed
                                   in dict PRODUCT_IDS) or serial number.
            id_kind (str): The kind of the given instr_id (listed in dict
                           ID_KINDS), defaults to device list index.
            contained_serial_no (int): Serial number of the desired device
                                       contained in the combined device.

        Returns:
            str: short device name
        """
        # translate id kind str to int
        id_kind_int = self.ID_KINDS[id_kind]

        # get length of device name string
        str_len = libtiepie.LstCbDevGetNameShort(id_kind_int, instr_id,
                                                 contained_serial_no, None, 0)

        # initialize mutable string buffer
        str_buffer = ctypes.create_string_buffer(str_len)

        # write the actual device name to the buffer
        libtiepie.LstCbDevGetNameShort(id_kind_int, instr_id,
                                       contained_serial_no,
                                       str_buffer, str_len)

        # convert to a normal python string
        dev_name = str_buffer.value.decode('utf-8')

        return dev_name

    def get_device_name_shortest_cb(self, instr_id, contained_serial_no,
                                    id_kind="index"):
        """Get the shortest name of a contained device in a combined
        instrument.

        Args:
            instr_id (int or str): Device list index, product ID (listed
                                   in dict PRODUCT_IDS) or serial number.
            id_kind (str): The kind of the given instr_id (listed in dict
                           ID_KINDS), defaults to device list index.
            contained_serial_no (int): Serial number of the desired device
                                       contained in the combined device.
        Returns:
            str: shortest device name of the device
        """
        # translate id kind str to int
        id_kind_int = self.ID_KINDS[id_kind]

        # get length of device name string
        str_len = libtiepie.LstCbDevGetNameShortest(id_kind_int, instr_id,
                                                    contained_serial_no,
                                                    None, 0)

        # initialize mutable string buffer
        str_buffer = ctypes.create_string_buffer(str_len)

        # write the actual device name to the buffer
        libtiepie.LstCbDevGetNameShortest(id_kind_int, instr_id,
                                          contained_serial_no, str_buffer,
                                          str_len)

        # convert to a normal python string
        dev_name = str_buffer.value.decode('utf-8')

        return dev_name

    def get_driver_version_cb(self, instr_id, contained_serial_no,
                              id_kind="index"):
        """Get the driver version of a contained device in a combined
        instrument.

        Not tested.

        Args:
            instr_id (int): Device list index, product ID
                            (listed in dict PRODUCT_IDS) or serial number.
            id_kind (str): the kind of the given instr_id
                           (listed in dict ID_KINDS), defaults to device list
                           index.
            contained_serial_no (int): Serial number of the desired device
                                       contained in the combined device.
        Returns:
            str: driver version of the device
        """
        # translate id kind str to int
        id_kind_int = self.ID_KINDS[id_kind]

        # Translate instr_id to int, if it is a product id str
        if id_kind == "product id":
            instr_id_int = self.PRODUCT_IDS[instr_id]
        else:
            instr_id_int = instr_id
        # get the driver version
        driver_version = libtiepie.LstCbDevGetDriverVersion(id_kind_int,
                                                            instr_id_int,
                                                            contained_serial_no)

        return version_to_str(driver_version)

    def get_firmware_version_cb(self, instr_id, contained_serial_no,
                             id_kind="index"):
        """Get the firmware version of a contained device in a combined
        instrument.

        Args:
            instr_id (int): Device list index, product ID
                            (listed in dict PRODUCT_IDS) or serial number.
            id_kind (str): the kind of the given instr_id
                           (listed in dict ID_KINDS), defaults to device list
                           index.
            contained_serial_no (int): Serial number of the desired device
                                       contained in the combined device.

        Returns:
            str: firmware version of the device
        """
        # translate id kind str to int
        id_kind_int = self.ID_KINDS[id_kind]

        # Translate instr_id to int, if it is a product id str
        if id_kind == "product id":
            instr_id_int = self.PRODUCT_IDS[instr_id]
        else:
            instr_id_int = instr_id
        # get the firmware version
        firmware_version = libtiepie.LstCbDevGetFirmwareVersion(id_kind_int,
                                                                instr_id_int,
                                                                contained_serial_no)

        return version_to_str(firmware_version)

    def get_calibration_date_cb(self, instr_id, contained_serial_no,
                                id_kind="index"):
        """Get the calibration date of the device.

        Args:
            instr_id (int): Device list index, product ID
                            (listed in dict PRODUCT_IDS) or serial number
            id_kind (str): the kind of the given instr_id
                           (listed in dict ID_KINDS), defaults to device list
                           index
            contained_serial_no (int): Serial number of the desired device
                                       contained in the combined device.

        Returns:
            :py:class:`datetime.date`: calibration date of the device
        """
        # translate id kind str to int
        id_kind_int = self.ID_KINDS[id_kind]

        # Translate instr_id to int, if it is a product id str
        if id_kind == "product id":
            instr_id_int = self.PRODUCT_IDS[instr_id]
        else:
            instr_id_int = instr_id

        # get the calibration date
        raw_date = libtiepie.LstCbDevGetCalibrationDate(id_kind_int,
                                                        instr_id_int,
                                                        contained_serial_no)

        split_date = date(raw_date >> 16, (raw_date >> 8) & 0xff, raw_date & 0xff)
        return split_date


def version_to_str(raw_version):
    """Convert a raw version int in TiePie's format to a nice string.

    Args:
        raw_version (int): concatenated version numbers as int in TiePie's format

    Returns:
        str: version number as str
    """
    return '.'.join([str((raw_version >> (idx * 16)) & 0xffff) for idx in range(3, -1, -1)])


# Instantiate class to make it available via import. Thus only one instance 
# exists (singleton design pattern).
device_list = DeviceList()
